from __future__ import annotations
from shapely import Point, LineString
from shapely.ops import nearest_points
from typing import Tuple, Union, Literal
from isr_matcher._constants._parse import db_km_to_km, km_to_db_km
from isr_matcher.ops._functions import split_line_at_point, only_km_db, only_m_db
import numpy as np
from isr_matcher._constants.logging import setup_logger
import logging


# Create logger for the current module
setup_logger()
logger = logging.getLogger(__name__)


# class Kilometrage
class Kilometrage:
    """Class that manages and stores kilometrage information for TrackSegment instances.

    Parameter
    ---------
    km_lines: list[list[km_lines]]
        A list of kilometrage lines, one for each direction. A kilometrage line is represented as sequential list of (geometrically connected) LineString instances.

    Attributes
    ----------
    n_lines: int
        The number of kilometratge lines.
    length_m: float
        The geometric length of kilometrage lines in meter.
    length_km: float
        The geometric length of kilometrage lines in kilometer.
    length_nominal_m: float
        The nominal geometric length in meter for all directions.
    length_nominal_km: float
        The nominal geometric length in kilometer for all directions.
    line: list[list[float]]
        All kilometrage lines as a list which contains a list of LineString objects for each existing direction of kilometrage.
    points: list[list[Point]]
        All kilometrage points as a list which contains a list of Point objects for each existing direction of kilometrage.
    values: list[list[str]]
        All kilometrage values as a list which contains a list of values for each existing direction of kilometrage.
    lengths: list[list[float]]
        All nominal geometric lengths of kilometrage segments in meter
    """

    def __init__(self, km_lines: list[list[LineString]], enhance_kilometrage: bool = True):
        """
        Initialize a Kilometrage object.

        Parameters:
        -----------
        km_lines : list[list[LineString]]
            List of LineStrings representing the kilometrage line(s).
        enhance_kilometrage : bool, optional
            Flag indicating whether to enhance kilometrage computation, by default True.

        Raises:
        -------
        ValueError
            If the number of lines is not exactly one or two.
        TypeError
            If the first element of the first list has a wrong type.
        """
        # verify input
        if len(km_lines) != 1 and len(km_lines) != 2:
            raise ValueError(
                f'list lines has {len(km_lines)} elements. It must contain exactly one or two LineStrings, representing the kilometrage line(s).'
            )

        if not isinstance(km_lines[0][0], LineString):
            raise TypeError(
                f'First elements of first list has wrong type: {type(km_lines[0][0])}. List lines must contain exactly one or two lists containing LineStrings, representing the kilometrage line(s).'
            )

        # set attributes
        self._n_km_lines: int = len(km_lines)  # number of kilometrage lines
        self._lines: list[list[LineString]] = km_lines  # kilometrage lines
        self._km_points: list[list[Point]] = [[] for _ in range(self._n_km_lines)]  # kilometrage points (coordinates)
        self._km_values: list[list[str]] = [[] for _ in range(self._n_km_lines)]  # kilometrage values at above points
        self._km_lengths: list[list[float]] = [
            [] for _ in range(self._n_km_lines)
        ]  # geometric length of segments between kilometrage points
        self._km_lengths_type: list[list[str]] = [
            [] for _ in range(self._n_km_lines)
        ]  # type of length ('nominal', 'computed')
        self._offset_m: list[list[float]] = [[] for _ in range(self._n_km_lines)]  # offsets at start of km segments
        self._previous_km_error: list[float] = [
            0.0 for _ in range(self._n_km_lines)
        ]  # last kilometrage segment length error of previous track segment
        self._enhance_kilometrage = enhance_kilometrage
        self._has_km_info = False  # flag whether km info from geo-strecken has been integrated
        self._offset_from_start_m: float = 0
        self.compute_geometric_length()  # computes geometric length of kilometrage line(s)
        if enhance_kilometrage:
            self.compute_geometric_offset()

    @property
    def n_lines(self) -> int:
        """Returns number of kilometratge lines."""
        return self._n_km_lines

    @property
    def length_m(self) -> list[list[float]]:
        """Returns the geometric length of kilometrage lines in meter."""
        return self._length

    @property
    def length_km(self) -> list[list[float]]:
        """Returns the geometric length of kilometrage lines in kilometer."""
        return [l / 1000 for l in self._length]

    @property
    def length_nominal_m(self) -> list[list[float]]:
        """Returns the nominal geometric length in meter for all directions."""
        return [np.sum([length for length in self._km_lengths[i]]) for i in range(self.n_lines)]

    @property
    def length_nominal_km(self) -> list[list[float]]:
        """Returns the nominal geometric length in kilometer for all directions."""
        return [np.sum([length / 1000 for length in self._km_lengths[i]]) for i in range(self.n_lines)]

    def line(self, direction: int = 0) -> list[LineString]:
        """Returns kilometrage line as list of lines for given direction.

        Parameter
        ---------
        direction: int
            The direction for which the kilometrage line(s) are returned.

        Returns
        -------
        line: The kilometrage line for the given directon.
        """
        if self.n_lines == 1:
            return self.lines[0]
        elif self.n_lines == 2:
            if direction == 1:
                return self.lines[0]
            elif direction == 2:
                return self.lines[1]
            else:
                msg = f'Received direction={direction} for track segment with two kilometrage lines. Set direction=1 for direction track or direction=2 for opposite track kilometrage.'
                raise ValueError(msg)
        else:
            raise NotImplementedError('Case not possible')

    @property
    def lines(self) -> list[list[LineString]]:
        """Returns all kilometrage lines as a list which contains a list of LineString objects for each existing direction of kilometrage."""
        return self._lines

    @property
    def points(self) -> list[list[Point]]:
        """Returns all kilometrage points as a list which contains a list of Point objects for each existing direction of kilometrage."""
        return self._km_points

    @property
    def values(self) -> list[list[str]]:
        """Returns all kilometrage values as a list which contains a list of values for each existing direction of kilometrage."""
        return self._km_values

    @property
    def lengths(self) -> list[list[float]]:
        """Returns all nominal geometric lengths of kilometrage segments in meter. Nominal geometric lengths are returned as a list which contains a list of lengths (m) for each existing direction of kilometrage."""
        return self._km_lengths

    @property
    def lengths_type(self) -> list[list[str]]:
        """Returns all lengths types (nominal, computed) of kilometrage segments in meter. Nominal geometric lengths are returned as a list which contains a list of lengths (m) for each existing direction of kilometrage."""
        return self._km_lengths_type

    @property
    def offset_m(self) -> list[list[float]]:
        """Returns all offsets of kilometrage segments in meter. Offsets are returned as a list which contains a list of offsets (m) for each existing direction of kilometrage."""
        return self._offset_m

    def points_in_direction(self, direction: int) -> list[Point]:
        """Returns kilometrage points for one direction. If track segment has only one km line for both directions, parameter direction is ignored."""
        if self.n_lines == 1:
            return self.points[0]
        elif self.n_lines == 2:
            if direction == 1:
                return self.points[0]
            elif direction == 2:
                return self.points[1]
            else:
                msg = f'Received direction={direction} for track segment with two kilometrage lines. Set direction=1 for direction track or direction=2 for opposite track kilometrage.'
                raise ValueError(msg)
        else:
            raise NotImplementedError('Case not possible')

    def values_in_direction(self, direction: int) -> list[str]:
        """Returns kilometrage values for one direction. If track segment has only one km line for both directions, parameter direction is ignored."""
        if self.n_lines == 1:
            return self.values[0]
        elif self.n_lines == 2:
            if direction == 1:
                return self.values[0]
            elif direction == 2:
                return self.values[1]
            else:
                msg = f'Received direction={direction} for track segment with two kilometrage lines. Set direction=1 for direction track or direction=2 for opposite track kilometrage.'
                raise ValueError(msg)
        else:
            raise NotImplementedError('Case not possible')

    def lengths_in_direction(self, direction: int) -> list[float]:
        """Returns nominal geometric lengths for one direction. If track segment has only one km line for both directions, parameter direction is ignored."""
        if self.n_lines == 1:
            return self.lengths[0]
        elif self.n_lines == 2:
            if direction == 1:
                return self.lengths[0]
            elif direction == 2:
                return self.lengths[1]
            else:
                msg = f'Received direction={direction} for track segment with two kilometrage lines. Set direction=1 for direction track or direction=2 for opposite track kilometrage.'
                raise ValueError(msg)
        else:
            raise NotImplementedError('Case not possible')

    def lengths_type_in_direction(self, direction: int) -> list[str]:
        """Returns type of geometric lengths for one direction ('computed', 'nominal'). If track segment has only one km line for both directions, parameter direction is ignored."""
        if self.n_lines == 1:
            return self.lengths_type[0]
        elif self.n_lines == 2:
            if direction == 1:
                return self.lengths_type[0]
            elif direction == 2:
                return self.lengths_type[1]
            else:
                msg = f'Received direction={direction} for track segment with two kilometrage lines. Set direction=1 for direction track or direction=2 for opposite track kilometrage.'
                raise ValueError(msg)
        else:
            raise NotImplementedError('Case not possible')

    def offset_in_direction(self, direction: int) -> list[float]:
        """Returns offset at the start of each kilometrage segment."""
        if self.n_lines == 1:
            return self.offset_m[0]
        elif self.n_lines == 2:
            if direction == 1:
                return self.offset_m[0]
            elif direction == 2:
                return self.offset_m[1]
            else:
                msg = f'Received direction={direction} for track segment with two kilometrage lines. Set direction=1 for direction track or direction=2 for opposite track kilometrage.'
                raise ValueError(msg)
        else:
            raise NotImplementedError('Case not possible')

    def _set_lengths_for_direction(self, lengths: list[float], direction: int):
        """Private method to set geometric lengths for one direction."""
        if self.n_lines == 1 and direction == 0:
            self._km_lengths[0] = lengths
        elif self.n_lines == 2 and direction == 1:
            self._km_lengths[0] = lengths
        elif self.n_lines == 2 and direction == 2:
            self._km_lengths[1] = lengths
        else:
            raise ValueError(
                f'Received direction = {direction} but track segment has {self.n_lines} direction(s). If one direction exists, direction must be 0. If two directions exist, direction must be 1 or 2'
            )

    def _set_lengths_type_for_direction(self, lengths_type: list[str], direction: int):
        """Private method to set geometric lengths type for one direction."""
        if self.n_lines == 1 and direction == 0:
            self._km_lengths_type[0] = lengths_type
        elif self.n_lines == 2 and direction == 1:
            self._km_lengths_type[0] = lengths_type
        elif self.n_lines == 2 and direction == 2:
            self._km_lengths_type[1] = lengths_type
        else:
            raise ValueError(
                f'Received direction = {direction} but track segment has {self.n_lines} direction(s). If one direction exists, direction must be 0. If two directions exist, direction must be 1 or 2'
            )

    def _set_km_lines_for_direction(self, km_lines: list[LineString], direction: int):
        """Private method to set kilometrage lines for one direction."""
        if self.n_lines == 1 and direction == 0:
            self._lines[0] = km_lines
        elif self.n_lines == 2 and direction == 1:
            self._lines[0] = km_lines
        elif self.n_lines == 2 and direction == 2:
            self._lines[1] = km_lines
        else:
            raise ValueError(
                f'Received direction = {direction} but track segment has {self.n_lines} direction(s). If one direction exists, direction must be 0. If two directions exist, direction must be 1 or 2'
            )

    def _set_pre_values(self, points: list[Point], values: list[str]):
        """Method to set kilometrage values for one direction."""
        sorted_index = self.argsort_km_values(values)
        self._pre_km_values = [values[i] for i in sorted_index]
        self._pre_km_points = [points[i] for i in sorted_index]

    def _pre_values(self) -> Tuple[list[str], list[Point]]:
        """Method which returns pre km values and points."""
        return self._pre_km_values, self._pre_km_points

    def add_km_point(self, point: Point, value: str, length: float, direction: int):
        """Method to add kilometrage points, values and geometric lengths to respective attributes for direction. Note that values are appended.

        Parameter
        ---------
        point: Point
            A new point to add to the kilometrage line (coordinates)
        value: str
            The kilometrage value at the new point as string (DB format)
        length: float
            The nominal geometric length from new point to next point (meter)
        direction: int
            The direction. For one kilometrage line (one direction) directionc can only be 0. For two kilometrage lines (two directions), it can be 1 or 2.
        """

        # get key (index of list)
        if (self.n_lines == 1 and direction == 0) or (self.n_lines == 2 and direction == 1):
            key = 0
        elif self.n_lines == 2 and direction == 2:
            key = 1
        else:
            raise ValueError(
                f'Received direction = {direction} but track segment has {self.n_lines} direction(s). If one direction exists, direction must be 0. If two directions exist, direction must be 1 or 2'
            )

        # append
        self._km_points[key].append(point)
        self._km_values[key].append(value)
        self._km_lengths[key].append(length)

        # sort and set
        sorted_index = self.argsort_km_values(self._km_values[key])

        self._km_points[key] = [self._km_points[key][i] for i in sorted_index]
        self._km_values[key] = [self._km_values[key][i] for i in sorted_index]
        self._km_lengths[key] = [self._km_lengths[key][i] for i in sorted_index]

    @staticmethod
    def argsort_km_values(values: list[str]) -> list[int]:
        """Sort kilometrage values in DB format.

        The function sorts kilometrage values in db format (strings) while considering overlength notation.

        Parameter
        ---------
        values: list[str]
            List of kilometrage values in db format (e.g. '3.6 + 53').
        """
        sorted_index = np.argsort([float(val.split('+')[0].replace(',', '.')) for val in values])
        for i in range(len(sorted_index) - 1):
            # sets correct order for same km values by comparing meter
            if only_km_db(values[sorted_index[i]]) == only_km_db(values[sorted_index[i + 1]]) and only_m_db(
                values[sorted_index[i]]
            ) > only_m_db(values[sorted_index[i + 1]]):
                temp = sorted_index[i]
                sorted_index[i] = sorted_index[i + 1]
                sorted_index[i + 1] = temp

            # same as above but now meter include overlength
            elif (
                only_km_db(values[sorted_index[i]]) == only_km_db(values[sorted_index[i + 1]])
                and only_m_db(values[sorted_index[i]]) == only_m_db(values[sorted_index[i + 1]])
                and only_m_db(values[sorted_index[i]], with_overlength=True)
                > only_m_db(values[sorted_index[i + 1]], with_overlength=True)
            ):
                temp = sorted_index[i]
                sorted_index[i] = sorted_index[i + 1]
                sorted_index[i + 1] = temp

        return sorted_index

    def compute_geometric_length(self):
        """Computes the geometric length in meter for each direction of track segment and set attribute."""

        # compute lengths for each direction
        if self.n_lines == 1:
            km_lines = self.lines[0]
            length = [np.sum([line.length for line in km_lines])]

        elif self.n_lines == 2:
            km_lines = self.lines
            length = [np.sum([line.length for line in km_lines[i]]) for i in range(2)]

        else:
            raise NotImplementedError(f'self.n_km_lines = {self.n_lines}. Possible values should only be 1 or 2.')

        # set attribute
        self._length = length

    def compute_nominal_length(self):
        """Computes the nominal geometric length in meter for each direction of track segment and set attribute."""
        directions = [0] if self.n_lines == 1 else [1, 2]
        for direction in directions:
            km_lengths = self.lengths_in_direction(direction=direction)
            km_values = self.values_in_direction(direction=direction)
            km_lengths_type = self.lengths_type_in_direction(direction=direction)

            for i in range(len(km_lengths)):
                if km_lengths_type[i] == 'computed':
                    km_lengths[i] = (db_km_to_km(km_values[i + 1]) - db_km_to_km(km_values[i])) * 1000

            dir_key = 0 if direction in [0, 1] else 1
            self._km_lengths[dir_key] = km_lengths

    def compute_geometric_offset(self):
        """Computes the geometric offset (overlength, underlength) for each kilometrage segment based on official geometric lengths."""

        geometric_offset_m: list[float] = [0]
        # compute lengths for each direction
        if self.n_lines == 1:
            km_lines = self.lines[0]
            lengths = self.lengths_in_direction(direction=0)
            lengths_types = self.lengths_type_in_direction(direction=0)

            segment_lengths = [line.length for line in km_lines]

            for ol, sl, lt in zip(lengths[:-1], segment_lengths[:-1], lengths_types[:-1]):
                difference_m = sl - ol
                if abs(difference_m) > 3:
                    offset_m = difference_m

                else:
                    offset_m = 0

                # append current offset
                geometric_offset_m.append(offset_m)

            # transfer over-/underlength to following segments

            for i in range(len(geometric_offset_m) - 1):
                if lengths_types[i + 1] == 'computed' and any([lt == 'nominal' for lt in lengths_types]):
                    geometric_offset_m[i + 1] += geometric_offset_m[i]

                else:
                    geometric_offset_m[i + 1] = 0

            geometric_offset_list = [geometric_offset_m]

        elif self.n_lines == 2:
            geometric_offset_list = []
            for direction in [1, 2]:
                geometric_offset_m: list[float] = [0]

                km_lines = self.lines[direction - 1]
                lengths = self.lengths_in_direction(direction=direction)
                lengths_types = self.lengths_type_in_direction(direction=direction)
                segment_lengths = [line.length for line in km_lines]

                for ol, sl, lt in zip(lengths[:-1], segment_lengths[:-1], lengths_types[:-1]):
                    # if lt == 'nominal':
                    difference_m = sl - ol
                    if abs(difference_m) > 3:
                        offset_m = difference_m

                    else:
                        offset_m = 0

                    # append current offset
                    geometric_offset_m.append(offset_m)

                # transfer over-/underlength to following segments
                for i in range(len(geometric_offset_m) - 1):
                    if lengths_types[i + 1] == 'computed' and any([lt == 'nominal' for lt in lengths_types]):
                        geometric_offset_m[i + 1] += geometric_offset_m[i]

                    else:
                        geometric_offset_m[i + 1] = 0

                geometric_offset_list.append(geometric_offset_m)

        else:
            raise NotImplementedError(f'self.n_km_lines = {self.n_lines}. Possible values should only be 1 or 2.')

        # set attribute
        self._offset_m = geometric_offset_list

    def correct_point_positions(self):
        """Corrects points positions by nominal segment lengths."""
        directions = [0] if self._n_km_lines == 1 else [1, 2]
        direction_key = 0
        for direction in directions:
            if len(directions) == 2:
                direction_key = direction - 1
            km_points = self.points_in_direction(direction=direction)
            km_lines = self.lines[direction_key]
            if len(km_points) - len(km_lines) > 1:
                km_points = km_points[: 1 + len(km_lines) - len(km_points)]

            # if difference between nominal geometric length and geometric length aligns for two neigbouring segments
            # move km point accordingly
            for i in range(1, len(km_points) - 1):
                # get values
                km_points = self.points_in_direction(direction=direction)
                km_lengths = self.lengths_in_direction(direction=direction)
                km_lines = self.lines[direction_key]
                km_segment_lengths = [line.length for line in km_lines]
                km_lengths_types = self.lengths_type_in_direction(direction=direction)

                if abs(km_lengths[i - 1]) < 1e-6 or abs(km_lengths[i]) < 1e-6:
                    # skip km jumps
                    continue

                diff_1 = km_lengths[i - 1] - km_segment_lengths[i - 1]
                diff_2 = km_lengths[i] - km_segment_lengths[i]

                if (
                    diff_1 * diff_2 < 0 and min([abs(diff_1), abs(diff_2)]) > 1.5
                ):  # errors must have opposite sign and must be greater than threshold
                    if (
                        km_lengths_types[i - 1] == km_lengths_types[i] == 'nominal'
                    ):  # or km_lengths_types[i-1] == km_lengths_types[i] == 'computed':
                        if abs(diff_1) > abs(diff_2):
                            if diff_2 < 0:
                                new_km_point = km_lines[i].interpolate(-diff_2)
                                move_right = True
                            else:
                                new_km_point = km_lines[i - 1].interpolate(km_segment_lengths[i - 1] - diff_2)
                                move_right = False
                        else:
                            # abs(diff_1) < abs(diff_2)
                            if diff_1 > 0:
                                new_km_point = km_lines[i].interpolate(diff_1)
                                move_right = True
                            else:
                                new_km_point = km_lines[i - 1].interpolate(km_lengths[i - 1])
                                move_right = False

                        if move_right == True:
                            new_line_1, new_line_2 = split_line_at_point(km_lines[i], new_km_point)
                            new_km_line_1 = LineString(list(km_lines[i - 1].coords) + list(new_line_1.coords))
                            new_km_line_2 = LineString(new_line_2.coords)

                        else:  # move left
                            new_line_1, new_line_2 = split_line_at_point(km_lines[i - 1], new_km_point)
                            new_km_line_1 = LineString(new_line_1.coords)
                            new_km_line_2 = LineString(list(new_line_2.coords) + list(km_lines[i].coords))

                        # remove original line
                        rem = km_lines.pop(i)
                        rem = km_lines.pop(i - 1)

                        # add splitted km lines to existing km lines
                        km_lines.insert(i - 1, new_km_line_1)
                        km_lines.insert(i, new_km_line_2)

                        km_points[i] = new_km_point

                self._km_points[direction_key] = km_points
                self._lines[direction_key] = km_lines

    def km(self, point: Point, direction: int = 0, return_type: Literal['db', 'float'] = 'float') -> Union[float, str]:
        """Returns the kilometrage value for a given point and direction.

        The computation of the kilometrage value is based on the geometry of the kilometrage lines as well as the kilometrage points and values of the Kilometrage instance.

        Parameter
        ---------
        point: Point
            The UTM coordinates (horizontal, vertical) for which kilometrage will be computed. Point will be projected on kilometrage line without consideration of the points distance. Must be passed as shapely Point instance.
        direction:
            The direction for which the kilometer is computed.
        return_type: Literal['db', 'float']
            Controls the return type of the function. If set to 'db' returns the kilometer as string in db-format (e.g. '3.6 + 53'). If set to 'float', returns kilometer as float.

        Returns
        -------
        km: float | str
            The computed kilometrage value. The type of returned variable depends on parameter, see 'return_type'.
        """
        # number of kilometrage lines
        n_lines = self.n_lines
        # error message
        direction_error_str = f'Invalid value for direciont: {direction}. There exist two kilometrage lines, one for each direction. Set direction=1 or direction=2.'

        # verify input direction and get kilometrage lines and values if legit direction
        if n_lines == 1:
            if direction != 0:
                # direction doesn't matter
                logger.debug(
                    f'Received direction={direction}. Ignoring direction as there exists only on kilometrage line.'
                )

            km_lines = self.lines[0]
            km_values = self.values_in_direction(direction=0)

        elif n_lines == 2:
            # two directions
            if direction == 0:
                raise ValueError(direction_error_str)

            elif direction == 1 or direction == 2:
                km_lines = self.lines[direction - 1]
                km_values = self.values_in_direction(direction=direction)

            else:
                raise ValueError(direction_error_str)

        else:
            raise NotImplementedError(f'self.n_km_lines = {self.n_lines}. Possible values should only be 1 or 2.')

        # get km_line segment closest to point
        distances = [point.distance(km_line_seg) for km_line_seg in km_lines]
        min_index = np.argmin(distances)
        km_segment = km_lines[min_index]

        # kilometer value at start of kilometrage segment
        km_segment_start = db_km_to_km(km_values[min_index])
        km_segment_end = db_km_to_km(km_values[min_index + 1])

        # coordinates of segment line
        # km_points = [Point(coord) for coord in km_segment.coords]

        # project point on km segment
        point_on_kilometrage = nearest_points(km_segment, point)[0]

        # if only km information from ISR (and not from GEO-STRECKE dataset)
        if self._enhance_kilometrage == False:
            km = km_segment_start + km_segment.project(point_on_kilometrage) / 1000
            if return_type == 'float':
                return round(km, 4)
            elif return_type == 'db':
                return km_to_db_km(km)

        else:  # complex computation of kilometrage
            # get kilometrage
            km_percentage = km_segment.project(point_on_kilometrage, normalized=True)

            # always one km line (min_index) less than km values
            if min_index > len(km_values) - 2:
                nominal_length_m = (db_km_to_km(km_values[-1]) - db_km_to_km(km_values[min_index])) * 1000
            else:
                nominal_length_m = (db_km_to_km(km_values[min_index + 1]) - db_km_to_km(km_values[min_index])) * 1000

            # precompute km line error (difference between geometric length of km line(s) and nominal geometric length)
            #            key = 0 if direction < 2 else 1
            #            km_length_km = np.sum(self.length_m[key]) / 1000

            # current and next kilometrage value (always at start of segment)
            curr_km_value = km_values[min_index]
            next_km_value = km_values[min_index + 1]

            km_length_type = self.lengths_type_in_direction(direction=direction)[min_index]
            offset_m_list = self.offset_in_direction(direction=direction)

            # distance based on km values at start and end of segment
            nominal_segment_length_km = db_km_to_km(next_km_value) - db_km_to_km(curr_km_value)

            # nominal geometric length of segment
            nominal_geometric_length_m = self.lengths_in_direction(direction=direction)[min_index]
            nominal_length_m = round(1000 * nominal_segment_length_km)

            # compute difference between the geometric length of kilometrage line and the nominal geometric length of line (from db-geo-strecken dataset)
            segment_error_m = km_segment.length - nominal_geometric_length_m
            official_diff = db_km_to_km(km_values[-1]) - db_km_to_km(km_values[0])

            km_points_ = self.points_in_direction(direction=direction)
            if min_index < len(km_points_) - 1 and km_points_[min_index + 1].distance(point_on_kilometrage) < 3:
                # if km is almost at end of segment, set km value to end of segment
                km = db_km_to_km(km_values[min_index + 1])

            elif nominal_geometric_length_m < 1e-6:
                # km jump: every km value inside the jump is set to km value after km jump
                # note: segment should be very short (~ 1m)
                if min_index < len(km_lines) - 1:
                    km_segment_end = db_km_to_km(km_values[min_index + 1])
                    km = km_segment_end
                else:
                    km_segment_start = db_km_to_km(km_values[min_index])
                    km = km_segment_start
                # km_db = km_to_db_km(km)

            elif abs(segment_error_m) < 1:
                # difference between nominal geometric length and geometric length is small: scale to official length
                distance_from_start_km = km_percentage * (nominal_geometric_length_m / 1000)
                km = km_segment_start + distance_from_start_km

            else:
                # difference between offical length and geometric length is not small and positive: overlength
                # -> treat as overlength after offical length is reached
                # difference between offical length and geometric length is not small and negative: missing length
                # -> treat as km jump at end of segment

                # km line error of current kilometrage segment
                segment_error_m = km_segment.length - nominal_geometric_length_m

                if segment_error_m < 0:  # no km info from geo strecken
                    # segment error negative: missing length and no km info
                    # missing length at start: offset
                    # missing length at end: no offset
                    if (
                        self._offset_from_start_m != 0 and min_index == 0
                    ):  # offset from start is only considered for first segment
                        distance_from_start_km = (
                            km_percentage * (km_segment.length / 1000) + self._offset_from_start_m / 1000
                        )
                    elif self._has_km_info == False:
                        distance_from_start_km = km_percentage * (nominal_geometric_length_m / 1000)
                    elif km_length_type == 'computed':  # has_km_info is True
                        distance_from_start_km = km_percentage * (nominal_geometric_length_m / 1000)
                    else:
                        distance_from_start_km = km_percentage * (km_segment.length / 1000)

                elif segment_error_m > 0:
                    if self._offset_from_start_m != 0 and min_index == 0:  # final
                        distance_from_start_km = (
                            km_percentage * (km_segment.length / 1000) + self._offset_from_start_m / 1000
                        )
                    elif self._has_km_info == False:
                        # treat error as continously increasing with track length -> substract percentage of error
                        distance_from_start_km = km_percentage * (km_segment.length / 1000) - km_percentage * (
                            segment_error_m / 1000
                        )
                    elif km_length_type == 'computed':  # has_km_info is True
                        distance_from_start_km = km_percentage * (nominal_geometric_length_m / 1000)
                    else:
                        distance_from_start_km = km_percentage * (km_segment.length / 1000)
                else:
                    # distance_from_start_m = km_percentage * (km_segment.length / 1000)
                    raise NotImplementedError('Can never be accessed.')

                km = km_segment_start + distance_from_start_km

            offset_m = offset_m_list[min_index]
            km += offset_m / 1000

            # special formatting when km value (float) exceeds value at segment end
            if km > db_km_to_km(next_km_value) or only_km_db(km_to_db_km(km)) > only_km_db(next_km_value):
                # if km is greater than last km value of segment treat as overlength
                # overlength in db format is shown as '14,9 + 52 | +1000'
                curr_km, curr_meter = curr_km_value.split('+')

                if db_km_to_km(curr_km_value) >= db_km_to_km(next_km_value):
                    km_db = f'{curr_km} +{curr_meter} | +{round(1000*(km - db_km_to_km(next_km_value)) + nominal_length_m,1)}'
                else:
                    # meter_added = round(float(curr_meter) + nominal_length_m)
                    # if nominal_length_m < 0: nominal_length_m = 0
                    overlength_m = round(1000 * (km - db_km_to_km(next_km_value)))
                    temp_km = km - overlength_m / 1000
                    km_at_overlength_start = km_to_db_km(temp_km)
                    km_db = f'{km_at_overlength_start} | +{round(overlength_m, 1)}'

                # get kilometer and meter values seperately
                def to_km(db_string):
                    return float(db_string.split(' | ')[0].split('+')[0].replace(',', '.').strip())

                def to_m(db_string):
                    return float(db_string.split(' | ')[0].split('+')[1].replace(',', '.').strip())

                next_only_km = to_km(next_km_value)
                curr_only_km = to_km(km_db)
                curr_only_m = to_m(km_db)

                # current km greater than next km
                if next_only_km < curr_only_km:
                    # compute difference and overlength
                    diff_only_km = curr_only_km - next_only_km
                    new_only_km = next_only_km
                    new_only_m = round(curr_only_m + diff_only_km * 1000, 1)

                    if '|' in km_db:
                        overlength_str = km_db.split(' | ')[1]
                        if (
                            new_only_km == next_only_km
                            and (float(overlength_str.replace('+', '')) + new_only_m) / 1000 < diff_only_km
                        ):
                            km_db = f'{new_only_km} +{float(overlength_str.replace("+","")) + new_only_m}'
                        else:
                            km_db = f'{new_only_km} +{new_only_m} | {overlength_str}'
                    else:
                        km_db = f'{new_only_km} +{new_only_m}'

            elif km < db_km_to_km(curr_km_value):
                diff_m = round(km - db_km_to_km(curr_km_value), 3) * 1000
                km_db = f'{curr_km_value} | +{diff_m}'

            else:
                km_db = km_to_db_km(km)

            # return type
            if return_type == 'db':
                return km_db
            elif return_type == 'float':
                return km
            else:
                raise ValueError(f'Unkown return_type: {return_type}')

    def integrate_kilometrage_points(
        self,
        values_list: Tuple[str, list[Point], Union[Literal['before', 'after'], None], float],
        direction: int,
        boundary_kms: Tuple[str, str],
        last_val: bool = False,
        first_val: bool = False,
    ):
        """Integrates official kilometrage point given by parameter 'values_list'. Must only be used by TrackSegment instances passing kilometrage information from data set db-geo-strecken, as it is a very specialized function. To simply set kilometrage points, use Kilometrage.add_km_point.

        The general idea is to split the kilometrage line into multiple segments, where one segment always runs from one km value (point) to the next km value (point).

        Parameters
        ----------
        values_list: Tuple[str, list[Point], Union[Literal['before', 'after'], None], float])
            - values_list[0]: str
                The kilometrage value in DB format as string.
            - values_list[1]: list[Point]
                The geometric point of the kilomatrage value. If this list contains more than one point, that means km value is outside the geometry of the track, so geometry of km_line will be expanded using those points.
            - value_list[2]: Union[Literal['before', 'after'], None]
                If values_list[1] has more than one point, this parameter indicates in which direction km line will be expanded
            - value_list[3]: float
                The official length in meter until the next kilometrage value.
        direction: int
            The direction for which kilomerage points will be added.
        """

        # extract values
        km_str = values_list[0]
        point_list = values_list[1]
        merge = values_list[2]
        nominal_geometric_length = values_list[3]

        # get km lines for direction
        km_lines = self.line(direction=direction)

        # if no expansions of km line (one point)
        if isinstance(merge, type(None)) or merge == 'end_point':
            assert len(point_list) == 1, f'Wrong number of points in list {len(point_list)} (== 1)'

            # nearest line to km point
            point = point_list[0]
            distances = [point.distance(line) for line in km_lines]
            min_index = np.argmin(distances)
            km_line = km_lines[min_index]

            # km values
            km_values = self.values_in_direction(direction=direction)
            km_points = self.points_in_direction(direction=direction)

            # get projected distances along kilometrage
            point_running_distance = km_line.project(point)
            km_points_running_distances = km_line.project(km_points)

            # dont add point if it doesnt fit with the logical order of already existing km points
            if not any(
                [
                    point_running_distance < km_point_rd
                    and db_km_to_km(km_str) > db_km_to_km(km_val)
                    and km_line.distance(km_point) < 10
                    for (km_point_rd, km_val, km_point) in zip(km_points_running_distances, km_values, km_points)
                ]
            ):
                # project km point on line
                point_on_km_line = nearest_points(point, km_line)[1]

                # set point
                self.add_km_point(
                    point=point_on_km_line, value=km_str, length=nominal_geometric_length, direction=direction
                )

                # check if point is at boundary of km line
                is_boundary_point = (point_on_km_line.distance(Point(km_line.coords[0])) < 1e-4) or (
                    point_on_km_line.distance(Point(km_line.coords[-1])) < 1e-4
                )

                # get projected distances
                p0_projected = km_line.project(point_list[0], normalized=True)
                p0_kmline = nearest_points(km_line, point_list[0])[0]

                # no boundary point -> splitting line possible
                if is_boundary_point == False:
                    # split line at point
                    km_line_1, km_line_2 = split_line_at_point(line=km_line, point=point_on_km_line)

                    # km values
                    km_values = self.values_in_direction(direction=direction)

                    # check if value is op from
                    if km_str.replace(',', '.').replace(' ', '') == boundary_kms[0].replace(',', '.').replace(' ', ''):
                        # if op from is no boundary point means it does not lie at start of km line
                        # therefore we need to compute a new value for the start of the km line
                        # based on its geometry and then set the computed value

                        # set point to start of km line
                        start_point = Point(km_lines[0].coords[0])
                        # index of op from
                        from_km_index = km_values.index(km_str)

                        # compute
                        km = (
                            db_km_to_km(km_str)
                            - km_line_1.length / 1000
                            - np.sum([line.length for line in km_lines[:from_km_index]]) / 1000
                        )
                        km_db = km_to_db_km(km)

                        # check if value or point already exists
                        km_value_exists = any([abs(km - db_km_to_km(km_)) * 1000 < 1 for km_ in km_values])
                        km_points = self.points_in_direction(direction=direction)
                        km_point_exists = any([start_point.distance(km_point) < 1e-4 for km_point in km_points])

                        # only add point if km value does not
                        # only split km line if km value does not exist but the respective point already exists
                        if km_value_exists == False and km_point_exists == False:
                            # remove original line
                            rem = km_lines.pop(min_index)
                            # add splitted km lines to existing km lines
                            km_lines.insert(min_index, km_line_1)
                            km_lines.insert(min_index + 1, km_line_2)
                            self.add_km_point(
                                point=start_point, value=km_db, length=km_line_1.length, direction=direction
                            )
                        elif km_value_exists == False and km_point_exists == True:
                            # remove original line
                            rem = km_lines.pop(min_index)
                            # add splitted km lines to existing km lines
                            km_lines.insert(min_index, km_line_1)
                            km_lines.insert(min_index + 1, km_line_2)

                    # check if value is op to
                    elif km_str.replace(',', '.').replace(' ', '') == boundary_kms[1].replace(',', '.').replace(
                        ' ', ''
                    ):
                        # if op to is no boundary point means it does not lie at end of km line
                        # therefore we need to compute a new value for the end of the km line
                        # based on its geometry and then set the computed value

                        # set end point
                        end_point = Point(km_lines[-1].coords[-1])

                        # compute km value
                        km = db_km_to_km(km_str) + km_line_2.length / 1000
                        km_db = km_to_db_km(km)

                        # check if km value or point already exists
                        km_value_exists = any([abs(km - db_km_to_km(km_)) * 1000 < 1 for km_ in km_values])
                        km_points = self.points_in_direction(direction=direction)
                        km_point_exists = any([end_point.distance(km_point) < 1e-4 for km_point in km_points])

                        # only add if value and point dont exist
                        if km_value_exists == False and km_point_exists == False:
                            # remove original line
                            rem = km_lines.pop(min_index)
                            # add splitted km lines to existing km lines
                            km_lines.insert(min_index, km_line_1)
                            km_lines.insert(min_index + 1, km_line_2)
                            self.add_km_point(
                                point=end_point, value=km_db, length=km_line_2.length, direction=direction
                            )
                        elif km_value_exists == False and km_point_exists == True:
                            # remove original line
                            rem = km_lines.pop(min_index)
                            # add splitted km lines to existing km lines
                            km_lines.insert(min_index, km_line_1)
                            km_lines.insert(min_index + 1, km_line_2)

                    else:
                        # remove original line
                        rem = km_lines.pop(min_index)
                        # add splitted km lines to existing km lines
                        km_lines.insert(min_index, km_line_1)
                        km_lines.insert(min_index + 1, km_line_2)
                        # can this be deleted ? Should never be entered
                        if (
                            first_val == True
                            and len(km_lines) == 2
                            and len(km_values) == 1
                            and (
                                km_str.replace(',', '.').replace(' ', '')
                                == boundary_kms[0].replace(',', '.').replace(' ', '')
                            )
                        ):
                            km_db = km_to_db_km(db_km_to_km(km_str) - km_lines[0].length / 1000)
                            self.add_km_point(
                                point=Point(km_lines[0].coords[0]),
                                value=km_db,
                                length=km_lines[0].length,
                                direction=direction,
                            )

                # get km values
                km_values = self.values_in_direction(direction=direction)

                # correction if something went wrong and one km_value was not set
                # check if a km point exists at the start and end of kilometrage
                # and if not, compute the km and it with the point
                if len(km_values) == len(km_lines) and last_val == True:
                    km_points = self.points_in_direction(direction=direction)
                    if not km_lines[0].project(km_points[0]) < 1e-3:
                        km_0 = km_to_db_km(db_km_to_km(km_values[0]) - km_lines[0].length / 1000)
                        self.add_km_point(
                            point=Point(km_lines[0].coords[0]),
                            value=km_0,
                            length=km_lines[0].length,
                            direction=direction,
                        )
                    elif not km_lines[-1].project(km_points[-1]) < 1e-3:
                        km_end = km_to_db_km(db_km_to_km(km_values[-1]) + km_lines[-1].length / 1000)
                        self.add_km_point(
                            point=Point(km_lines[-1].coords[-1]),
                            value=km_end,
                            length=km_lines[-1].length,
                            direction=direction,
                        )

                # set new km lines
                self._set_km_lines_for_direction(km_lines=km_lines, direction=direction)

        # else (serveral points)
        else:
            assert (merge == 'before') or (merge == 'after'), f'Wrong value for merge: {merge}'

            # nearest line to points
            distances = [np.mean(line.distance(point_list)) for line in km_lines]
            min_index = np.argmin(distances)
            km_line = km_lines[min_index]

            # compute projected distances along km line of point
            p0_projected = km_line.project(point_list[0], normalized=True)
            p0_kmline = nearest_points(km_line, point_list[0])[0]

            # checks if the merge flag may be erronous by verifying if the point already lies on the km line (instead of being some distance away from it)
            if (
                len(point_list) == 1
                and ((p0_projected > 1e-4) and (p0_projected) < 1 - 1e-4)
                and p0_kmline.distance(point_list[0]) < 1e-3
            ):
                # wrong merge (point is already on line, no need to merge)
                split_point = point_list[0]
                km_line_1, km_line_2 = split_line_at_point(line=km_line, point=split_point)

                # remove original line
                rem = km_lines.pop(min_index)

                # add splitted km lines to existing km lines
                km_lines.insert(min_index, km_line_1)
                km_lines.insert(min_index + 1, km_line_2)

                # set new km lines and add point
                self._set_km_lines_for_direction(km_lines=km_lines, direction=direction)
                self.add_km_point(point=split_point, value=km_str, length=km_line_2.length, direction=direction)

                # add value at end of rail (bigger than km_to)
                km_end_str = km_to_db_km(db_km_to_km(km_str) + km_line_2.length / 1000)

                self.add_km_point(point=Point(km_line_2.coords[-1]), value=km_end_str, length=0, direction=direction)

                if merge == 'before' and first_val == True and Point(km_lines[0].coords[0]).distance(split_point) > 0:
                    km_values = self.values_in_direction(direction=direction)
                    prev_km_val = km_to_db_km(db_km_to_km(km_values[0]) - km_lines[0].length / 1000)
                    self.add_km_point(
                        point=Point(km_line.coords[0]), value=prev_km_val, length=km_line.length, direction=direction
                    )

            # add points to existing km line according to value_list[2] (merge: before, after, None) (raise error if None)
            elif merge == 'before':
                # get km values and points
                km_values = self.values_in_direction(direction=direction)
                km_points = self.points_in_direction(direction=direction)
                # next km value that is greater than given kilometer
                next_km_val = [km_val for km_val in km_values if db_km_to_km(km_val) > db_km_to_km(km_str)]

                if len(next_km_val) == 0:
                    # no next km because it is the first point added
                    # have to use pre values (are set beforehand in TrackSegment.integrate_kilometrage_points)
                    km_values, km_points = self._pre_values()
                    next_km_val = [
                        km_val
                        for km_val in km_values
                        if (db_km_to_km(km_val) > db_km_to_km(km_str) or only_km_db(km_val) > only_km_db(km_str))
                        and not (only_km_db(km_val) < only_km_db(km_str))
                    ]

                if len(next_km_val) > 0:
                    # merge
                    index = km_lines.index(km_line)
                    rem = km_lines.pop(index)
                    km_line = LineString(point_list + list(km_line.coords))
                    km_lines.insert(index, km_line)

                    # get next km point
                    index = km_values.index(next_km_val[0])
                    next_km_point = km_points[index]

                    # project next km point on km line if it is not alredy on the line
                    if next_km_point.distance(km_line) > 1e-6:
                        next_km_point = nearest_points(km_line, next_km_point)[0]

                    # check if next point is at the end of km line -> split not possible
                    distance_percent = km_line.project(next_km_point, normalized=True)
                    if not ((distance_percent == 1)):
                        # split
                        km_line_1, km_line_2 = split_line_at_point(line=km_line, point=next_km_point)

                        # remove original line
                        rem = km_lines.pop(min_index)

                        # add splitted km lines to existing km lines
                        km_lines.insert(min_index, km_line_1)
                        km_lines.insert(min_index + 1, km_line_2)

                    # first point
                    point = point_list[0]
                    self._set_km_lines_for_direction(km_lines=km_lines, direction=direction)
                    self.add_km_point(point=point, value=km_str, length=nominal_geometric_length, direction=direction)

            else:
                # merge after
                km_line = LineString(list(km_line.coords) + point_list)

                # set split point
                split_km_point = point_list[0]

                # verify if boundary
                is_boundary_point = (split_km_point.distance(Point(km_line.coords[0])) < 1e-4) or (
                    split_km_point.distance(Point(km_line.coords[-1])) < 1e-4
                )
                if is_boundary_point == False:
                    # split at next km point
                    km_line_1, km_line_2 = split_line_at_point(km_line, split_km_point)

                    # remove original line
                    rem = km_lines.pop(min_index)

                    # add splitted km lines to existing km lines
                    km_lines.insert(min_index, km_line_1)
                    km_lines.insert(min_index + 1, km_line_2)
                    # first point
                    point = point_list[0]
                    # set
                    self._set_km_lines_for_direction(km_lines=km_lines, direction=direction)
                    self.add_km_point(point=point, value=km_str, length=nominal_geometric_length, direction=direction)

                else:
                    km_lines[-1] = km_line
                    self._set_km_lines_for_direction(km_lines=km_lines, direction=direction)
                    self.add_km_point(
                        point=point_list[-1], value=km_str, length=nominal_geometric_length, direction=direction
                    )

        self._has_km_info = True



# Script that contains functions for converting strings from json response (see utils._ISR_PROPERTIES).

from typing import Any, Tuple, Union, overload
import numpy as np
from dataclasses import dataclass
from isr_matcher._constants._filters import ISR_EXCEPTIONAL_STATION_NAMES

# data classes used for type hints


@dataclass
class dtype:
    dtype: Any


# tuple collection of integer and float types

INT_TYPES = (int, np.integer, np.int16, np.int32, np.int64)
FLOAT_TYPES = (int, np.integer, np.int16, np.int32, np.int64, float, np.float16, np.float32, np.float64)


# conversion functions


def identity(obj: Any) -> Any:
    """Returns the input."""
    return obj


def str_to_int(string: str) -> int:
    """Converts input string to integer."""
    return int(string)


@overload
def str_to_float(string: str) -> float:
    ...


@overload
def str_to_float(string: None) -> None:
    ...


def str_to_float(string: Union[str, None]) -> Union[float, None]:
    # TODO: refactor this function
    """Converts input string from ISR properties to float."""
    try:
        to_replace = ['Bremsweg', 'A', 'm', '(ET 420)']
        for replace in to_replace:
            string = string.replace(replace, '')
        return float(string.strip().replace(',', '.').replace(' ', ''))
    except:
        return string


def incline_to_array(string: str) -> np.ndarray:
    """Converts string with inclination profile to numpy array."""
    # TODO
    return string


def db_km_to_km(string: str) -> float:
    """Converts string with kilometrage from db format to km."""
    if '|' in string:
        string, overlength = string.split(' | ')
        overlength = float(overlength.replace('+', ''))
    else:
        overlength = 0
    km, m = string.split('+')
    if overlength:
        m = float(m)
        m += overlength
    return float(km.replace(',', '.')) + float(m) / 1000.0


def km_to_db_km(value: float) -> str:
    """Converts kilometer values to db format."""
    km = round(int(value * 10) / 10, 1)
    m = round(int(value * 10000) / 10000 - km, 3) * 1000
    return f'{km}'.replace('.', ',') + f' + {int(m)}'


def db_km_alt_to_km(number: int | str) -> float:
    """Converts string with kilometrage from alternative db format to km."""
    if isinstance(number, str):
        number = int(number)
    km = round((number - 1e8) / 1e5, 1)
    m = round(number - 1e8 - km * 1e5)
    return km + m / 1000


def str_to_lat_lon(string: str) -> Tuple[float, float]:
    """Converts string to latitude and longitude coordinate."""
    if 'auf Anfrage' in string:
        return (-1.0, -1.0)
    elif '+,' in string:
        return (-1.0, -1.0)
    else:
        lat_string, lon_string = string.replace('+', '').split(',')
    return (float(lat_string), float(lon_string))


def extract_station_names(
    track_name_string: str, split_only: bool = False, is_op_name: bool = False
) -> Tuple[str, str] | str:
    """TODO"""

    exceptional_names = ISR_EXCEPTIONAL_STATION_NAMES

    exceptional_name_list = [name for name in exceptional_names if name in track_name_string]

    if split_only:
        if len(exceptional_name_list) > 0:
            d = {}
            for i, exceptional_name in enumerate(exceptional_name_list):
                d[str((i + 1) * 10000)] = exceptional_name
                track_name_string = track_name_string.replace(exceptional_name, str((i + 1) * 10000))
            from_name, to_name = track_name_string.split(' - ')
            for i in range(len(exceptional_name_list)):
                from_name = from_name.replace(str((i + 1) * 10000), d[str((i + 1) * 10000)])
                to_name = to_name.replace(str((i + 1) * 10000), d[str((i + 1) * 10000)])
        else:
            from_name, to_name = track_name_string.split(' - ')

    else:
        if len(exceptional_name_list) > 0:
            for exceptional_name in exceptional_name_list:
                # 1
                if ' - ' in exceptional_name:
                    wild_card_name = exceptional_name.replace(' - ', '*').replace('-', '*')
                else:
                    wild_card_name = None

                # 2
                if '   ' in exceptional_name:
                    if wild_card_name:
                        name_split = wild_card_name.split('   ')
                    else:
                        name_split = exceptional_name.split('   ')
                    station_base_name = name_split[0]
                    additional_name = name_split[-1].strip()
                    wild_card_name = '*' + station_base_name + '*' + additional_name + '*'
                else:
                    wild_card_name = '*' + exceptional_name.replace('  ', ' ').replace(' ', '*') + '*'

                # 3
                if 'ß' in wild_card_name:
                    wild_card_name = wild_card_name.replace('ß', '*')

                track_name_string = track_name_string.replace(exceptional_name, wild_card_name)

                if is_op_name == True:
                    return track_name_string

        from_name, to_name = track_name_string.split(' - ')

    return from_name, to_name



"""Script that contains dictionaries with explanations and transformations for track properties."""

from isr_matcher._constants._parse import (
    identity,
    str_to_int,
    str_to_float,
    incline_to_array,
    db_km_to_km,
    str_to_lat_lon,
    db_km_alt_to_km,
)

# fmt: off

##### ISR - Infrastrukturregister der Deutschen Bahn

### Abkürzungen

# ALG - Allgemein
# ISR - Infrastrukturregister
# INF - Infrastruktur
# BET - Betrieb
# LST - Zugsteuerung, Zugsicherung und Signalgebung
# ENE - Energie

### STRECKENABSCHNITTE 

# TODO: Check if integer properties should use identity (int -> int) or str_to_int (str -> int). (Depends on json string)

ISR_PROPERTIES_TRACK_SEGMENTS = {
    "ALG_BESONDERE_PRUEF": ["Streckenkilometer bestimmter Stellen, die besondere Prüfung erfordern", "Lage bestimmter Stellen, die wegen Abweichungen von den oben genannten Lichtraumprofilen besondere Prüfungen erfordern", identity],
    "ALG_BESONDERE_PRUEF_DOK": ["Dokument mit den Querschnitten der Stellen, die besondere Prüfung erfordern", "Elektronisches Dokument, das vom Infrastrukturbetreiber zur Verfügung gestellt und von der ERA gespeichert wird, mit den Querschnitten der Stellen, die wegen Abweichungen von den oben genannten Lichtraumprofilen besondere Prüfungen erfordern. Dem Dokument mit dem Querschnitt können gegebenenfalls Hinweise für die Prüfung der bestimmten Stellen beigefügt werden.", identity],
    "ALG_BSTGZIELHOEHE": ["Bahnsteigzielhöhe [mm]", "Soll-Abstand zwischen der Bahnsteigoberkante und der Lauffläche des benachbarten Gleises. Die angegebenen Bahnsteighöhen sind Systemhöhen. Diese beinhalten auch Übergangsflächen zwischen verschiedenen Systemhöhen und beziehen sich immer auf die Soll-Gleislage. Die tatsächliche Bahnsteighöhe kann auf Grund der Ist-Gleislage und Instandhaltungstoleranzen von der Systemhöhe abweichen.", str_to_float],
    "ALG_DBNETZ_STRECKE": [],
    "ALG_GUE_KORR_7": ["Angabe, ob die Strecke dem Schienengüterverkehrskorridor 7 gemäß EU-Verordnung (EU)913/2010 zugeordnet ist", "Ein Güterverkehrskorridor (englisch Rail Freight Corridor, RFC) ist eine ausgewiesene Eisenbahnstrecke zwischen zwei oder mehr Staaten der Europäischen Union, die zwei oder mehr Bahnhöfe entlang einer Hauptroute miteinander verbindet. Sie werden auch SGV-Korridore genannt.", identity],
    "ALG_GUE_KORR_9": ["Angabe, ob die Strecke dem Schienengüterverkehrskorridor 9 gemäß EU-Verordnung (EU)913/2010 zugeordnet ist", "Ein Güterverkehrskorridor (englisch Rail Freight Corridor, RFC) ist eine ausgewiesene Eisenbahnstrecke zwischen zwei oder mehr Staaten der Europäischen Union, die zwei oder mehr Bahnhöfe entlang einer Hauptroute miteinander verbindet. Sie werden auch SGV-Korridore genannt.", identity],
    "ALG_GUE_KORR_A": ["Angabe, ob die Strecke dem Schienengüterverkehrskorridor 1 gemäß EU-Verordnung (EU)913/2010 zugeordnet ist", "Ein Güterverkehrskorridor (englisch Rail Freight Corridor, RFC) ist eine ausgewiesene Eisenbahnstrecke zwischen zwei oder mehr Staaten der Europäischen Union, die zwei oder mehr Bahnhöfe entlang einer Hauptroute miteinander verbindet. Sie werden auch SGV-Korridore genannt.", identity],
    "ALG_GUE_KORR_B": ["Angabe, ob die Strecke dem Schienengüterverkehrskorridor 3 gemäß EU-Verordnung (EU)913/2010 zugeordnet ist", "Ein Güterverkehrskorridor (englisch Rail Freight Corridor, RFC) ist eine ausgewiesene Eisenbahnstrecke zwischen zwei oder mehr Staaten der Europäischen Union, die zwei oder mehr Bahnhöfe entlang einer Hauptroute miteinander verbindet. Sie werden auch SGV-Korridore genannt.", identity],
    "ALG_GUE_KORR_D": ["Angabe, ob die Strecke dem Schienengüterverkehrskorridor 4 gemäß EU-Verordnung (EU)913/2010 zugeordnet ist", "Ein Güterverkehrskorridor (englisch Rail Freight Corridor, RFC) ist eine ausgewiesene Eisenbahnstrecke zwischen zwei oder mehr Staaten der Europäischen Union, die zwei oder mehr Bahnhöfe entlang einer Hauptroute miteinander verbindet. Sie werden auch SGV-Korridore genannt.", identity],
    "ALG_GUE_KORR_F": ["Angabe, ob die Strecke dem Schienengüterverkehrskorridor 8 gemäß EU-Verordnung (EU)913/2010 zugeordnet ist", "Ein Güterverkehrskorridor (englisch Rail Freight Corridor, RFC) ist eine ausgewiesene Eisenbahnstrecke zwischen zwei oder mehr Staaten der Europäischen Union, die zwei oder mehr Bahnhöfe entlang einer Hauptroute miteinander verbindet. Sie werden auch SGV-Korridore genannt.", identity],
    "ALG_IM_CODE": ["Infrastrukturbetreiber Code - Kennung des Infrastrukturbetreibers", "Numerische Codierung des Infrastrukturbetreibers gemäß UIC-Liste (verfügbar unter http://www.uic.org -> Company Codes) z. B. 0080 für die DB Netz AG.", identity],
    "ALG_INFRA_BETR": ["Infrastrukturbetreiber", "Infrastrukturbetreiber bezeichnet eine Einrichtung oder ein Unternehmen, die bzw. das insbesondere für die Einrichtung und Unterhaltung der Fahrwege der Eisenbahn zuständig ist.", identity],
    "ALG_INT_LIRA_PROFIL": ["Interoperables Lichtraumprofil", "Lichtraumprofile GA, GB, GC G1, DE3 gemäß europäischer Norm (EN 15273-3:2013).", identity],
    "ALG_KV_PROFIL": ["KV Kodifizierung", "Kodierung für den kombinierten Verkehr mit Wechselbehälter gemäß UIC-Merkblatt 596-6. Angabe nur auf Strecken wo auch kombinierter Verkehr tatsächlich gefahren wird.", identity],
    "ALG_LAENGE_ABSCHNITT": ["Länge des Streckenabschnittes [km]", "Länge der Strecke zwischen den Betriebsstellen am Beginn und Ende des Streckenabschnitts.", str_to_float],
    "ALG_MIN_LIRA_PROFIL": ["Multinationales Lichtraumprofil", "Multinationales Lichtraumprofil oder internationales Lichtraumprofil außer GA, GB, GC, G1, DE3, S, IRL1 gemäß europäischer Norm (in Deutschland nur G2)", identity],
    "ALG_NAT_LIRA_PROFIL": ["Nationales Lichtraumprofil", "Inländisches Lichtraumprofil gemäß europäischer Norm oder anderes lokales Lichtraumprofil (in Deutschland zurzeit DE1 und DE2). Die fahrzeugbezogenen Streckenfreigaben werden in den Anlagen 5.x zu den Grundsätzen des ISR veröffentlicht.", identity],
    "ALG_SPEZ_INFO": ["Spezifische Informationen", "Relevante Informationen des Infrastrukturbetreibers in Bezug auf die Trassierung.", identity],
    "ALG_STAAT": ["Staat", "Mitgliedstaat der europäischen Union (+UK +NO +CH).", identity],
    "ALG_STRECKENKLASSE": ["Streckenklasse", "Angabe der maximalen Radsatzlast und der zulässigen Last je Längeneinheit. Die Strecken der DB Netz AG werden in die Streckenklassen A – D4 nach DIN EN 15528 eingeteilt. Zusätzlich gelten folgende nationale Erweiterungen: https://fahrweg.dbnetze.com/resource/blob/1357560/5ce6d02c8c5ed5da47f7d065a269e61c/allgemein_grundsaetze_isr-data.pdf, Abschn. 2.1.1., Streckenklasse.", identity],
    "ALG_TEN_GIS_ID": ["TEN-GIS ID", "Angabe der GIS ID des Abschnitts der TEN-V-Datenbank, zu dem der Streckenabschnitt gehört.", identity],
    "ALG_TEN_GUETER": ["TSI Streckenkategorie Güterverkehr", "Die TSI-Streckenkategorie ergibt sich aus einer Kombination so genannter Verkehrscodes (Traffic Code). Die Einteilung werden durch folgende Leistungskennwerte bestimmt: Lichtraumprofil - Achsfahrmasse - Streckengeschwindigkeit - Zuglänge.", identity],
    "ALG_TEN_KLASSIFIZIERUNG": ["TEN Klassifizierung", "Angabe des Teils des transeuropäischen Netzes, zu dem die Strecke gehört nach TSI Infrastruktur (VERORDNUNG (EU) Nr. 1299/2014) gültig seit 01.01.2015.", identity],
    "ALG_TEN_PERSON": ["TSI Streckenkategorie Personenverkehr", "Die TSI-Streckenkategorie ergibt sich aus einer Kombination so genannter Verkehrscodes (Traffic Code). Die Einteilung werden durch folgende Leistungskennwerte bestimmt: Lichtraumprofil - Achsfahrmasse - Streckengeschwindigkeit - Bahnsteignutzlänge.", identity],
    "ALG_TEN_REGPERSON": ["TSI Streckenkategorie regionaler Personenverkehr", "TSI-Streckenkategorie für den regionalen Personenverkehr, wir ebenfalls durch die Leistungskennwerte Lichtraumprofil, Achsfahrmasse, Streckengeschwindigkeit und Bahnsteignutzlänge bestimmt.", identity],
    "ALG_VERKEHRSART": ["Verkehrsart", "Angabe der Verkehrsart (Pz, Gz, Mischverkehr, S-Bahn), 'Pz', 'Gz' oder 'S-Bahn' bedeutet, dass die jeweils anderen Verkehrsarten ausgeschlossen sind.", identity],
    "BET_BETRIEBSVERFAHREN": ["Betriebsverfahren", "Angabe des anzuwendenden Betriebsverfahren zum Verkehren von Zügen, Rangierfahrten und anderem (z. B. Betriebsverfahren nach Richtlinie 408, Zugleitbetrieb nach Richtlinie 436, 438 oder FV-NE, Signalisierter Zugleitbetrieb nach Richtlinie 437).", identity],
    "BET_EINSCHR_DOK": [],
    "BET_GESCHWINDIGKEIT": ["Geschwindigkeit [km/h]", "Zulässige betriebliche Höchstgeschwindigkeit auf der Strecke in Kilometer/Stunde.", str_to_float],
    "BET_NOTBREMS_UEBERBRUECK": ["Notbremsüberbrückung", "Angabe ob auf dem Streckenabschnitt, eine Notbremsüberbrückungseinrichtung erforderlich ist. Nach der EBA-Richtlinie 'Anforderung des Brand- und Katastrophenschutzes an den Bau und den Betrieb von Eisenbahntunneln (Stand 01.07.2008) gelten für TEN-Strecken die Vorschriften der TSI. Außerhalb des Geltungsbereiches des TEN und bei Ausnahmen zur TSI müssen die Notbremsen von allen in Reisezügen eingestellten Fahrzeugen, die lange und sehr lange Tunnel befahren, so beschaffen sein, dass eine durch Reisende eingeleitete Notbremsung bis zum Verlassen des Tunnels aufgehoben werden kann.", identity],
    "BET_OEFFNUNGSZEITEN": ["Streckenöffnungszeiten", "Informationen zu Öffnungszeiten der Strecken sind als Sachdaten zu den Betriebsstellen im ISR hinterlegt.", identity],
    "BET_PFV": ["Personenfernverkehr", "Angabe ob auf einem Streckenabschnitt Züge des Personenfernverkehrs verkehren.", identity],
    "DET_KAPAZITAETSBINDUNG": [],
    "ENE_ANF_STROMABN": ["Anforderung bzgl. der Zahl der ausgefahrenen Stromabnehmer und deren Abstand zueinander, bei gegebener Geschwindigkeit", "Angabe der zulässigen Höchstzahl der ausgefahrenen Stromabnehmer je Zug und des Mindestabstands der Mittellinien benachbarter Stromabnehmerwippen in Metern bei vorgegebener Geschwindigkeit. Zugelassen ist im Netz der DB ein Betrieb mit maximal zwei gehobenen Stromabnehmern in einem Abstand von x < 35 m bzw. x > 85 m.", identity],
    "ENE_AUTO_STROMABN_SENK": ["Automatische Stromabnehmer-Senkeinrichtung erforderlich", "Angabe, ob am Fahrzeug eine automatische Absenkeinrichtung vorhanden sein muss.", identity],   
    "ENE_BEGRENZ_BORD": ["Strombegrenzung an Bord erforderlich", "Angabe, ob eine fahrzeugseitige Strom- oder Leistungsbegrenzungsfunktion erforderlich ist.", identity],       
    "ENE_EGPRUEF_DOK": ["EG-Prüfung TSI ENE", "Eindeutige Nummer der EG-Erklärung gemäß den Formatvorgaben in 'Document aboutpractical arrangements for transmitting interoperability documents'.", identity],      
    "ENE_MAXSTROM_GZ": ["Höchster Zugstrom (Güterzug) [A]", "Angabe der maximal zulässigen Stromaufnahme der Güterzüge in Ampere (A). Je nach Zugsicherungssystem kann der angegebene Wert nicht vollständig ausgenutzt werden.", str_to_float],
    "ENE_MAXSTROM_PZ": ["Höchster Zugstrom (Personenzug) [A]","Angabe der maximal zulässigen Stromaufnahme der Personenzüge in Ampere (A). Je nach Zugsicherungssystem kann der angegebene Wert nicht vollständig ausgenutzt werden.", str_to_float],
    "ENE_MAXSTROM_STILL": ["Maximale Stromaufnahme bei Stillstand je Stromabnehmer [A]", "Angabe der maximal zulässigen Stromaufnahme der Züge bei Stillstand in Ampere (A). Für DC-Systeme überwiegend in den Grenzbereichen, eventuell auch Auskunft in den Grenzbetriebsvereinbarungen beachten.", str_to_float],
    "ENE_MAX_FAHRDRAHTHOEHE": ["Maximale Fahrdrahthöhe [mm]", "Angabe der maximalen Fahrdrahthöhe in Millimetern.", str_to_float],
    "ENE_MIN_FAHRDRAHTHOEHE": ["Minimale Fahrdrahthöhe [mm]", "Angabe der Mindestfahrdrahthöhe in Millimetern. Nach EBO Anlage 3 Abs. 3 liegt die Mindestfahrdrahthöhe bei 4,95m. Ausnahmen davon werden in der Liste sowie in der Übersichtskarte als Anlage 6 veröffentlicht. Die Daten zur freien Strecke finden sich zudem in der interaktiven Karte des ISR wieder.", str_to_float],
    "ENE_MITTL_KONTAKTKRAFT": ["Mittlere Kontaktkraft - Kurve", "Angabe der Kurve zur Bestimmung der zulässigen Kontaktkraft (AC C, AC C1, AC C2, DC 1,5 kV, DC 3 kV). Zur Bestimmung der mittleren Kontaktkraft siehe https://eur-lex.europa.eu/legal-content/DE/TXT/HTML/?uri=CELEX:32008D0284&from=PL#d1e260-1-1, Abschn. 4.2.15.", identity],
    "ENE_NUTZBREMS_ERL": ["Erlaubnis für Nutzbremsung vorhanden", "Angabe, ob Nutzbremsung erlaubt ist oder nicht. Bei Wechselstrom 15kV 16,7Hz und 25kV 50Hz immer erlaubt im Oberleitungsnetz der DB. Ebenso. im Oberleitungsnetz der DB, bei Gleichstrom 1,5 kV 3000A und 3kV 2400A. Für die S-Bahn Hamburg und Berlin kann keine Aussage getroffen werden.", identity],
    "ENE_PHASENTRENNSTR": ["Phasentrennstrecken (Schutzstrecken)", "Angabe, ob Phasentrennung vorhanden ist.", identity],
    "ENE_PHASENTRENNSTRSPEC": ["Phasentrennstrecken - Spezifikation", "Angabe mehrerer erforderlicher Daten zur Phasentrennung. Aus programmtechnischen Gründen erfolgt die Anzeige nur in Englisch. Bsp: ''length 144 + switch off breaker Y + lower pantograph Y' (length: Länge des spannungslosen Abschnitts in Meter, switch off breaker Y/N: Hauptschalter ausschalten Ja/Nein, lower pantograph Y/N: Stromabnehmer senken Ja/Nein).", identity],
    "ENE_SYSTEMTRENNSTR": ["Systemtrennstrecken", "Angabe, ob eine Systemtrennung vorhanden ist. Sind nur im Bereich von Grenzbahnhöfen und/oder grenzüberschreitenden Strecken vorhanden (Stromsystemwechsel). Regelungen sind in den Grenzvereinbarungen zu finden.", identity],
    "ENE_SYSTEMTRENNSTRSPEC": ["Systemtrennstrecken - Spezifikation", "Angabe mehrerer erforderlicher Daten zur Systemtrennung. Aus programmtechnischen Gründen erfolgt die Anzeige nur in Englisch. Bsp: 'length 144 + switch off breaker Y + lower pantograph Y + change supply system Y' (lenght: Länge des spannungslosen Abschnitts in Meter, switch off breaker Y/N: Hauptschalter ausschalten Ja/Nein, lower pantograph Y/N: Stromabnehmer senken Ja/Nein, change supply system Y/N: Wechsel des Versorgungssystems Ja/Nein).", identity],
    "ENE_TRAKT_STROMART": ["Stromversorgungssystem (Spannung und Frequenz)", "Angabe des Stromversorgungssystems (Nennspannung und –frequenz).", identity],
    "ENE_ZUL_SSWERKSTOFF": ["Zulässiger Schleifstückwerkstoff", "Angabe, welche Schleifstückwerkstoffe verwendet werden dürfen. In Deutschland reine Hartkohle und (ab 2017) imprägnierte Kohle mit Zusatz-Werkstoffen.", identity],
    "ENE_ZUL_STROMABN_BREITE": ["Zugelassene Stromabnehmerbreite [mm]", "Angabe von TSI-konformen Stromabnehmerwippen, die verwendet werden dürfen. Generell ist im 15kV-Netz der DB nur eine Stromabnehmerbreite von 1950mm zulässig. Liegt eine passende EG-Prüfung vor, kann auch eine Breite von 1600mm zugelassen sein.", str_to_float],
    "ID": ["Identifikatinsnummer", "Identifkitationsnummer im Infrastrukturregister.", identity],
    "INF_BUE_BESCHLEUNIGUNG": ["An Bahnübergängen erlaubte Beschleunigung [m/s^2]", "Grenzwert für die Beschleunigung des Zuges, falls er in der Nähe eines Bahnübergangs hält, in Meter pro Sekunde zum Quadrat.", str_to_float],
    "INF_EGPRUEF_DOK": ["EG Prüfung TSI INF", "Eindeutige Nummer der EG-Erklärung gemäß den Formatvorgaben in 'Document about practical arrangements for transmitting interoperability' (ERA/INF/10-2009/INT verfügbar auf der ERA-Webseite).", identity],
    "INF_ELEKTROMAG_BED": ["Bedingungen für den Einsatz von Magnetschienenbremsen", "Elektronisches Dokument, das vom Infrastrukturbetreiber zur Verfügung gestellt und von der ERA gespeichert wird, mit den Bedingungen für den Einsatz von Magnetschienenbremsen.", identity],
    "INF_ELEKTROMAG_ERL": ["Elektromagnetische Bremsen", "Angabe der Einschränkungen für den Einsatz von Magnetschienenbremsen.", identity],
    "INF_GLEISANZAHL": ["Gleis", "Angabe, ob die Streckenführung ein- oder zweigleisig ist. Bei zweigleisigen Strecken wird gemäß Ril 883.1000 zwischen dem Richtungsgleis und dem Gegengleis unterschieden. 'Bei zweigleisigen Strecken ist das Richtungsgleis das vom Streckenanfang in Richtung auf das Streckenende rechts gelegene Gleis. Das Gegenrichtungsgleis ist das in diesem Sinne links gelegene Gleis.' (Ril 883.1000)", identity],
    "INF_HOA_VORH": ["Streckenseitige Heißläuferortungsanlagen (HOA) vorhanden", "Die folgenden Alarmarten sind für alle Arten von HOA gültig: Warmalarm (Radsatzlager – Umgebungstemperatur), Heißalarm (Radsatzlager – Umgebungstemperatur), Differenzalarm (Differenz zwischen linkem und rechtem Achslager), Alarmgrenzwerte. Die folgenden Alarmgrenzwerte sind für alle Arten von HOA gültig: Warmalarm 70K, Heißalarm 100K, Differenzalarm 65K.", identity],
    "INF_HOECHSTHOEHE": ["Höchsthöhe [m]", "Höchster Punkt des Streckenabschnitts über Meereshöhe bezogen auf NAP (Normal Amsterdam’s Peil) in Metern.", str_to_float],
    "INF_HSLM": ["Konformität von Bauwerken mit dem dynamischen Lastmodell HSLM", "Für Streckenabschnitte mit zulässiger Höchstgeschwindigkeit von mindestens 200km/h. Informationen zum Verfahren zur Durchführung der Prüfung der dynamischen Kompatibilität.", identity],
    "INF_HSLM_DOK": ["Dokument mit dem Verfahren für statische und dynamische Streckenkompatibilitätsprüfungen", "Elektronisches Dokument, das vom Infrastrukturbetreiber zur Verfügung gestellt und von der ERA gespeichert wird, mit dem genauen Verfahren für statische und dynamische Prüfungen der Streckenkompatibilität oder einschlägigen Informationen für die Durchführung der Prüfungen an bestimmten Bauwerken.", identity],
    "INF_HSLM_KM": ["Streckenkilometrierung von Bauwerken, die besondere Prüfungen erfordern", "Lage von Bauwerken, die besondere Prüfungen erfordern.", identity],
    "INF_KOMM_SYSTEM": ["Kommunikationssystem", "Angaben des installierten Zugfunksystems (analoger Zugfunk, VZF 95 II, GSM-R oder GSM).", identity], 
    "INF_LEISE_STRECKE": ["Gehört zu einer leiseren Eisenbahnstrecke", "Gehört nach Artikel 5b der Verordnung (EU) Nr. 1304/2014 der Kommission zu einer 'leiseren Eisenbahnstrecke'.", identity],
    "INF_MAX_RAD_DURCHMESSER": ["Radmindestdurchmesser für stumpfe Kreuzungen [mm]", "Die maximal zulässige Herzstücklücke einer festen stumpfen Kreuzung beruht auf einem in Millimetern angegebenen Radmindestdurchmesser, auch nach Abnutzung, im Betrieb.", str_to_float],
    "INF_MAX_UEBERH_FEHLBETRAG": ["Maximaler Überhöhungsfehlbetrag [mm]", "Maximaler Überhöhungsfehlbetrag in Millimetern, definiert als Differenz zwischen der tatsächlichen Überhöhung und einer höheren Ausgleichserhöhung, für die die Strecke ausgelegt ist.", str_to_float],
    "INF_MAX_ZUGVERZOEGERUNG": ["Maximale Zugverzögerung [m/s^2]", "Grenzwert für die Gleislagestabilität in Längsrichtung, angegeben als höchstzulässige Zugverzögerung in Metern pro Sekunde zum Quadrat.", str_to_float],
    "INF_MIN_BOGENRADIUS": ["Kleinster Bogenhalbmesser [m]", "Halbmesser des kleinsten horizontalen Bogens des Gleises in Metern. Die dargestellten Mindestradien gelten richtungsbezogen auch für die Fahrt im durchgehenden Hauptgleis. Beim Befahren von Weichenverbindungen und anderen Gleisen können, gemäß EBO, auch Radien bis 150m auftreten.", str_to_float],
    "INF_NEIGETECHNIK": ["Neigetechnik", "Angabe der Zugbeeinflussungssysteme, welche auf diesen Streckenabschnitten (ZUB 262, ZUB 122 oder ZUB 122/ZUB) vorhanden sind.", identity],
    "INF_REGELSPURWEITE": ["Regelspurweite [mm]", "Die Regelspurweite ist das lichte Innenmaß zwischen den zwei Schienenköpfen. Die Regelspurweite beträgt in Deutschland und den meisten europäischen Ländern 1435 mm (auch: Normalspur).", str_to_float],
    "INF_SCHIENENNEIGUNG": ["Schienenneigung", "Neigung des Kopfes einer im Gleis verlegten Schiene gegenüber der Lauffläche (1:20 oder 1:40).", identity],
    "INF_SCHOTTER_VORH": ["Schotter vorhanden", "Angabe, ob das Phänomen Schotterflug beim Hochgeschwindigkeitsverkehr auftreten kann. Angaben nur bei Strecken mit Vmax > 250km/h.", identity],
    "INF_SEITENWINDREST": [],
    "INF_SPURKRANZ": ["Einsatz von Spurkranzschmierung untersagt", "Angabe, ob die Nutzung von fahrzeugseitigen Einrichtungen zur Spurkranzschmierung verboten ist.", identity],
    "INF_STEIGUNGSPROFIL": ["Steigungsprofil", "Abfolge der Steigungswerte und Angabe der Orte, an denen sich die Steigung ändert. Wird im Format [±NN.N] ([±NNNN.NNN]) Neigung (Ort) angegeben und beliebig oft wiederholt. Dabei ist der erste Ort die Lage der Startbetriebsstelle und damit der Beginn des ersten Neigungswertes. Die letzte Ortsangabe ist der Beginn des letzten Neigungswert, der dann in der Endbetriebsstelle endet. Der minimale Neigungswechsel beträgt 0,5 mm/m.", incline_to_array],
    "INF_STRECKENNEIGUNG": ["Streckenneigung", "Angabe der maximalen Neigung zwischen 2 Betriebsstellen in 5‰-Schritten zur besseren Darstellung als thematische Karte. Detaillierte Angaben zu Steigung und Gefälle auf dem Streckenabschnitt finden sich in den Angaben zum 'Steigungsprofil'.", identity],
    "INF_STRENG_KLIMA": ["Vorliegen strenger klimatischer Bedingungen", "Strenge oder normale klimatische Bedingung auf der Strecke gemäß europäischer Norm.", identity],
    "INF_STRIKT_LOKAL": ["Vorschriften und Einschränkungen strikt lokaler Art vorhanden", "Vorschriften und Einschränkungen strikt lokaler Art vorhanden.", identity],
    "INF_STRIKT_LOKAL_DOK": ["Dokument des Infrastrukturbetreibers zu Vorschriften und Einschränkungen strikt lokaler Art", "Elektronisches Dokument, das vom Infrastrukturbetreiber zur Verfügung gestellt und der ERA gespeichert wird, mit zusätzlichen Angaben zu den Vorschriften und Einschränkungen strikt lokaler Art.", identity],
    "INF_TEMP_SPANNE": ["Temperaturspanne [°C]", "Temperaturspanne für den uneingeschränkten Zugang zur Strecke gemäß europäischer Norm.", identity],
    "INF_TRAKTIONSART": ["Traktionsart", "Angabe über die Elektrifizierung (Oberleitung, Stromschiene) oder ohne Elektrifizierung.", identity],
    "INF_TSI_EINHALT_BETRIEBSWERT": ["TSI-Einhaltung der Betriebswerte für Weichen und Kreuzungen", "Weichen und Kreuzungen werden gemäß den in den TSI spezifizierten Betriebsgrenzwerten instandgehalten.", identity],
    "INF_WIRBELSTROM_BED": ["Bedingungen für den Einsatz von Wirbelstrombremsen", "Elektronisches Dokument, das vom Infrastrukturbetreiber zur Verfügung gestellt und von der ERA gespeichert wird, mit den Bedingungen für den Einsatz von Wirbelstrombremsen.", identity],
    "INF_WIRBELSTROM_ERL": ["Wirbelstrombremse", "Angabe der Einschränkungen für den Einsatz von Wirbelstrombremsen.", identity],   
    "ISEG_ID": [],
    "ISR_KM_BIS": ["Kilometer (bis) [km + m]", "Angaben der Kilometrierung (metergenau) der Betriebsstellen (meist Mitte Empfangsgebäude) am Ende eines Streckenabschnitts.", identity],
    "ISR_KM_BIS_I": ["Kilometer (bis) - DB Codierung", "Angaben der Kilometrierung (metergenau) der Betriebsstellen (meist Mitte Empfangsgebäude) am Ende eines Streckenabschnitts (DB Codierung).", str_to_int],
    "ISR_KM_VON": ["Kilometer (von) [km + m]", "Angaben der Kilometrierung (metergenau) der Betriebsstellen (meist Mitte Empfangsgebäude) am Anfang eines Streckenabschnitts.", identity],
    "ISR_KM_VON_I": ["Kilometer (von) - DB Codierung", "Angaben der Kilometrierung (metergenau) der Betriebsstellen (meist Mitte Empfangsgebäude) am Anfang eines Streckenabschnitts (DB Codierung).", str_to_int],
    "ISR_STEL_ID_BIS": ["Identifikationsnummer der Betriebsstelle am Streckenabschnittsende", "Identifikationsnummer im Infrastrukturregister der Betriebsstelle am Streckenabschnittsende.", identity],
    "ISR_STEL_ID_VON": ["Identifikationsnummer der Betriebsstelle am Streckenabschnittsanfang", "Identifikationsnummer im Infrastrukturregister der Betriebsstelle am Streckenabschnittsanfang.", identity],
    "ISR_STRECKE_VON_BIS": ["Streckenabschnitt", "Eindeutige Bezeichnungen der Betriebsstellen am Anfang und am Ende eines Streckenabschnitts. Dabei kann ein Streckenabschnitt nicht in derselben Betriebsstelle enden wie er gestartet ist.", identity],
    "ISR_STRE_NR": ["Strecken-Nr.", "Bezeichnung und Identifizierung für Eisenbahnstrecken in Deutschland.", identity],
    "JAHR": ["Jahr", "Angabe des Jahres.", identity],
    "LADE_ID": ["Lade ID", "Lade Identifkationsnummer im Infrastrukturregister.", identity],
    "LST_ARTEN_ZUGORTUNG": ["Art der Zugortungsanlage/Gleisfreimeldeeinrichtung", "Angabe der Arten von installierten Zugortungsanlagen/Gleisfreimeldeeinrichtungen.", identity],
    "LST_BESONDERE_PRUEF": ["Gleisstromkreise oder Achszähler, die besondere Prüfungen erfordern", "Angabe der Arten von Zugortungsanlagen/Gleisfreimeldeeinrichtungen, die besondere Prüfungen erfordern.", identity],
    "LST_BESONDERE_PRUEF_DOK": ["Dokument zur besonderen Prüfung von Gleisstromkreisen und Achszählern", "Elektronisches Dokument, das vom Infrastrukturbetreiber in zwei EU-Sprachen zur Verfügung gestellt und von der ERA gespeichert wird, mit dem genauen Verfahren für die besondere Prüfung der angegebenen Zugortungsanlagen/Gleisfreimeldeeinrichtungen.", identity],    
    "LST_BETRIEBSBESCHR": ["Betriebsbeschränkungen oder -bedingungen vorhanden", "Angabe, ob Beschränkungen oder Bedingungen aufgrund einer Teilkonformität mit der TSI ZZS — Verordnung (EU) 2016/919 der Kommission (5) vorhanden sind.", identity],
    "LST_DETAIL_ROAMING": ["Einzelheiten zum Roaming in öffentlichen Netzen", "Sofern Roaming in öffentlichen Netzen konfiguriert ist, wird hier angegeben, für welche Netze/ für welche Nutzer/für welche Gebiete.", identity],
    "LST_EGPRUEF_DOK": ["EG-Prüfung TSI ZZS", "Eindeutige Nummer der EG-Erklärung gemäß den Formatvorgaben in 'Document about practical arrangements for transmitting interoperability'.", identity],
    "LST_EINH_RST_SHUNT": ["TSI-Konformität der Vorschriften über Kombinationen von RST-Merkmalen mit Einfluss auf die Kurzschlussimpedanz", "Angabe, ob die Vorschriften mit der TSI im Einklang stehen.", identity],     
    "LST_ETCS_INFILL_ZUGANG": ["ETCS-Infill-Funktion für Streckenzugang notwendig", "Angabe, ob Infill aus Sicherheitsgründen für den Zugang erforderlich ist.", identity],     
    "LST_ETCS_LEVEL_STOER": ["ETCS-Level für gestörten Betrieb", "ERTMS/ETCS-Level hinsichtlich der streckenseitigen Ausrüstung für eingeschränkten Betrieb.", identity],  
    "LST_ETCS_LEVEL_VERS": ["ETCS Level", "ERTMS/ETCS-Anwendungsstufe hinsichtlich der streckenseitigen Ausrüstung (Level1, Level2 oder ohne ETCS).", identity],  
    "LST_ETCS_NATIONAL": ["Paket 44 der nationalen ETCS-Anwendung implementiert", "Angabe, ob Daten für nationale Anwendungen zwischen Gleis und Zug übertragen werden.", identity],  
    "LST_ETCS_SOFTWARE_VERS": ["ETCS SRS Version", "Angabe der streckenseitig installierten ETCS-Baseline Version (2.3.0 d, 3.3.0, 3.4.0).", identity],  
    "LST_ETCS_STRECKE_INFILL": ["Streckenseitig installierte ETCS-Infill-Funktion", "Information zu installierter streckenseitiger Ausrüstung, die Infill-Informationen mittels einer Schleife oder GSM-R für Installationen des Level 1 übertragen kann.", identity],   
    "LST_FERRO_RADWERKSTOFF": ["Ferromagnetische Eigenschaften des Radwerkstoffs vorgeschrieben TSI-konform", "Angabe, ob die Vorschriften mit der TSI im Einklang stehen.", identity],
    "LST_FUNKSYS_KLASSE_B": ["Andere Funksysteme installiert", "Angabe, ob streckenseitig im Normalbetrieb andere Funksysteme installiert sind.", identity],
    "LST_GPRS": ["GPRS für ETCS", "Angabe, ob GPRS für ETCS verwendet werden kann.", identity],
    "LST_GPRS_ANWEND": ["GPRS-Anwendungsbereich", "Angabe des Bereichs, in dem GPRS für das ETCS verwendet werden kann", identity],
    "LST_GRUPPE_555": ["Verwendung der Gruppe 555", "Angabe, ob die Gruppe 555 verwendet wird.", identity],
    "LST_GSMR_VERSION": ["GSM-R Version", "Nummern der streckenseitig installierten FRS- und SRS-Versionen des GSM-R.", identity],
    "LST_KEIN_GSMR": ["Keine GSM-R-Abdeckung", "Angabe, ob eine GSM-R-Abdeckung besteht oder nicht.", identity],
    "LST_LZB": ["LZB", "Angabe, ob der Streckenabschnitt, mit dem Zugsicherungssystem LZB ausgerüstet ist und mit welcher Version.", identity],
    "LST_MAX_SAND": ["Maximaler Sandausstoß [g]", "Angabe des maximalen Wertes des auf dem Gleis akzeptierten Sandausstoßes für einen Zeitraum von 30s in Gramm (500g oder 800g).", str_to_float],
    "LST_MINDESTZAHL_GSMR": ["Mindestanzahl aktiver GSM-R-Modems an Bord für die Datenübertragung", "Für einen reibungslosen Zugbetrieb empfohlene Zahl der Modems für die ETCS-Datenübertragung. Dies betrifft Kommunikationssitzungen mit Hilfe der Streckenzentrale (RBC).", identity],
    "LST_MIN_BREMSLEIST": ["Maximal geforderter Bremsweg [m]", "Angabe des maximalen Bremswegs (in Metern) eines Zuges für die Streckenhöchstgeschwindigkeit.", str_to_float],
    "LST_M_VERSION": ["ETCS M_Version", "ETCS M_Version gemäß SRS 7.5.1.9", identity],
    "LST_NETZMERKMALE": ["Zusätzliche Angaben zu den Netzmerkmalen", "Zusätzliche Angaben zu den Netzmerkmalen oder entsprechendes Dokument, das beim Infrastrukturbetreiber erhältlich ist und von der ERA gespeichert wird, z.B.: Interferenzniveau, das zur Empfehlung einer zusätzlichen fahrzeugseitigen Sicherung führt.", identity],    
    "LST_OPTION_GSMR": ["Optionale GSM-R-Funktionen", "Einsatz optionaler GSM-R-Funktionen, die den Betrieb auf der Strecke verbessern können.", identity],
    "LST_PZB": ["PZB", "Angabe, ob der Streckenabschnitt, mit dem Zugsicherungssystem PZB ausgerüstet ist.", identity],
    "LST_RANGIER": [],
    "LST_ROAMING_NETZ": ["Roaming in öffentlichen Netzen vorhanden", "Roaming in öffentlichen Netzen vorhanden.", identity],
    "LST_ROAMING_VEREINB": ["GSM-R-Netze, für die eine Roaming-Vereinbarung vorliegt", "Liste der GSM-R-Netze, für die eine Roaming-Vereinbarung vorliegt.", identity],
    "LST_SAND_CHAR": ["Vorschriften zur Sand Charakteristik sind TSI-konform", "Angabe, ob die Vorschriften mit der TSI im Einklang stehen.", identity], 
    "LST_SPURKRANZ": ["Existenz von Vorschriften zur fahrzeugseitigen Spurkranzschmierung", "Angabe, ob Vorschriften für die Aktivierung oder Deaktivierung der Spurkranzschmierung vorhanden sind.", identity], 
    "LST_SYSTEMKOMP": ["ETCS-Systemkompatibilität", "ETCS-Anforderungen zum Nachweis der technischen Kompatibilität.", identity],
    "LST_TSI_KONF_ABST": ["TSI-konformer Höchstabstand zwischen zwei aufeinanderfolgenden Achsen", "Angabe, ob der erforderliche Abstand mit der TSI im Einklang steht.", identity],
    "LST_TSI_KONF_IMP": ["TSI-konforme Höchstimpedanz zwischen gegenüberliegenden Rädern eines Radsatzes", "Angabe, ob die Vorschriften mit der TSI im Einklang stehen.", identity],
    "LST_UEBERG_FUNKSYS": ["Übergänge zwischen verschiedenen Funksystemen vorhanden", "Angabe, ob zwischen verschiedenen Systemen während der Fahrt umgeschaltet werden kann", identity],
    "LST_UEBERG_ZUGSICH": ["Übergänge zwischen verschiedenen Zugsicherungs-, Zugsteuerungs- und Warnsystem vorhanden", "Angabe, ob während der Fahrt zwischen verschiedenen Funksystemen umgeschaltet und das Kommunikationssystem ausgeschaltet werden kann.", identity],
    "LST_UNTERDR_SAND": ["Unterdrücken des Sandens durch den Triebfahrzeugführer vorgeschrieben", "Angabe, ob der Triebfahrzeugführer nach den Anweisungen des Infrastrukturbetreibers über die Möglichkeit verfügen muss, die Sandstreuanlagen zu aktivieren und zu deaktivieren.", identity],
    "LST_VERBUND_BREMS": ["Einhaltung der TSI bei Vorschriften zur Nutzung von Verbundstoffbremsklötzen", "Angabe, ob die Vorschriften mit der TSI im Einklang stehen.", identity],
    "LST_VOR_ELE_MAG": ["Vorschriften für die von Fahrzeugen ausgesendeten Magnetfelder vorhanden und TSI konform", "Angabe, ob Vorschriften vorhanden sind und mit der TSI im Einklang stehen.", identity],
    "LST_VOR_METALL_FREI": ["Vorschriften zu metallfreiem Raum in der Radumgebung TSI-konform", "Angabe, ob die Vorschriften mit der TSI im Einklang stehen.", identity],
    "LST_VOR_METALL_MASSE": ["Vorschriften für Metallmasse von Fahrzeugen TSI-konform", "Angabe, ob die Vorschriften mit der TSI im Einklang stehen.", identity],
    "LST_VOR_STROMRUECK": ["Vorschriften für Grenzen der Oberschwingungen des Traktionsstroms vorhanden und TSI-konform", "Angabe, ob Vorschriften vorhanden sind und mit der TSI im Einklang stehen.", identity],
    "LST_ZHOE_ABST_FOLG_ACHS": ["Zulässiger Höchstabstand zwischen zwei aufeinanderfolgenden Achsen falls nicht TSI konform [mm]", "Angabe des zulässigen Höchstabstands zwischen zwei aufeinanderfolgenden Achsen in Millimetern, falls nicht TSI-konform.", str_to_float],
    "LST_ZHOE_IMPEDANZ": ["Zulässige Höchstimpedanz zwischen gegenüberliegenden Rädern eines Radsatzes falls nicht TSI-konform [Ohm]", "Wert der zulässigen Höchstimpedanz in Ohm, falls nicht TSI-konform.", str_to_float],
    "LST_ZHOE_LAENG_BUG": ["Zulässiger Höchstabstand zwischen Fahrzeugende und erster Achse [mm]", "Angabe des Höchstabstands zwischen Fahrzeugende und erster Achse in Millimetern für beide Enden (vorderes und hinteres Ende) eines Fahrzeuges oder Zuges.", str_to_float],
    "LST_ZMAX_HOEHE_RADKR": ["Maximal zulässige Höhe des Spurkranzes [mm]", "Angabe der maximal zulässigen Höhe des Radkranzes in Millimetern.", str_to_float],
    "LST_ZMIN_ACHSLAST": ["Zulässige Mindestachslast [t]", "Angabe der zulässigen Mindestachslast in Tonnen.", str_to_float],
    "LST_ZMIN_ACHSLAST_FAHRZ": ["Zulässige Mindestradsatzlast je Fahrzeugklasse [t]", "Angabe der zulässigen Mindestachslast in Tonnen je nach Fahrzeugklasse.", identity],
    "LST_ZMI_ABST_E_L_ACHS": ["Zulässiger Mindestabstand zwischen erster und letzter Achse [mm]", "Angabe des zulässigen Mindestabstands zwischen erster und letzter Achse in Millimetern.", str_to_float],
    "LST_ZMI_ABST_FOLG_ACHS": ["Zulässiger Mindestabstand zwischen zwei aufeinanderfolgenden Achsen [mm]", "Angabe des zulässigen Mindestabstands zwischen zwei aufeinanderfolgenden Achsen in Millimetern.", str_to_float],
    "LST_ZMI_BREITE_RADKR": ["Zulässige Mindestbreite des Radkranzes [mm]", "Angabe der zulässigen Mindestbreite des Radkranzes in Millimetern.", str_to_float],
    "LST_ZMI_DICK_RADKR": ["Zulässige Mindestdicke des Spurkranzes [mm]", "Angabe der zulässigen Mindestdicke des Radkranzes in Millimetern.", str_to_float],
    "LST_ZMI_DURCH_RAD": ["Zulässiger Mindestdurchmesser des Rades [mm]", "Angabe des zulässigen Mindestraddurchmessers des Rades in Millimetern.", str_to_float],
    "LST_ZMI_HOEHE_RADKR": ["Zulässige Mindesthöhe des Spurkranzes [mm]", "Angabe der zulässigen Mindesthöhe des Radkranzes in Millimetern.", str_to_float],
    "LST_ZUGFUNK_DATEN": ["Kompatibilität des Zugfunksystems (Daten)", "Funkanforderungen zum Nachweis der technischen Kompatibilität (Daten).", identity],
    "LST_ZUGFUNK_SPRACHE": ["Kompatibilität des Zugfunksystems (Sprache)", "Funkanforderungen zum Nachweis der technischen Kompatibilität (Sprache).", identity],
    "LST_ZUGINTEGRITAET": ["Fahrzeugseitige Bestätigung der Zugintegrität für Streckenzugang notwendig", "Angabe, ob eine fahrzeugseitige Bestätigung der Zugintegrität aus Sicherheitsgründen für den Zugang zur Strecke erforderlich ist.", identity],
    "LST_ZUGORTUNG": ["TSI-konforme Zugortungsanlagen/Gleisfreimeldeeinrichtungen vorhanden", "Angabe, ob eine vollständig mit der TSI ZZS konforme Zugortungsanlage/Gleisfreimeldeeinrichtung installiert ist.", identity],
    "LST_ZUSATZ_IM": ["Zusätzliche Angaben beim Infrastrukturbetreiber erhältlich", "Angabe, ob zusätzliche Angaben gemäß Nummer 4.2.2.6.2 Unternummer 2 des Anhangs der Durchführungsverordnung (EU) 2019/773 beim Infrastrukturbetreiber erhältlich sind.", identity],
    "LST_ZUSATZ_IM_DOK": ["Dokument des Infrastrukturbetreibers zur Bremsleistung", "Elektronisches Dokument, das vom Infrastrukturbetreiber zur Verfügung gestellt und von der ERA gespeichert wird, mit zusätzlichen Angaben gemäß Nummer 4.2.2.6.2 Unternummer 2 des Anhangs der Durchführungsverordnung (EU) 2019/773 die beim Infrastrukturbetreiber erhältlich sind.", identity],
    "LST_ZZS_CLASSB_STOER": ["Andere Zugsicherungs-, Zugsteuerungs- und Warnsysteme für gestörten Betrieb", "Angabe, ob andere Systeme als ETCS für eingeschränkten Betrieb vorhanden sind.", identity],
    "SEG_ID": [],
}   


### BETRIEBSSTELLEN

ISR_PROPERTIES_OPERATIONAL_POINT = {
    "ALG_BAHNSTEIGE_DET": [],
    "ALG_BAHNUEBERGAENGE_DET": [],
    "ALG_BRUECKE_DET": [],
    "ALG_BSTEIG_BEREICH_PLAN": [],
    "ALG_DBNETZ_STRECKE": [],
    "ALG_DETAILPLAN_DOK": [],
    "ALG_EG_PRUEF_TSI_ENE": [],
    "ALG_EG_PRUEF_TSI_INF": [],
    "ALG_EG_PRUEF_TSI_LST": [],
    "ALG_EG_PRUEF_UEBERBL": [],
    "ALG_GEO_LAGE": ["Geografische Lage der Betriebsstelle [WGS84 - EPSG:4326]", "Geografische Koordinaten in Dezimalgrad, normalerweise in Bezug auf einen Punkt in der Mitte des Empfangsgebäudes", str_to_lat_lon],
    "ALG_GLEISE_DET": [],
    "ALG_GSMR_RAFU": [],
    "ALG_MOBILIPLAN_DOK": [],
    "ALG_REGELBAHNST_HOEHEN": ["Bahnsteighöhen [mm]", "Die angegebenen Bahnsteighöhen sind Systemhöhen. Diese beinhalten auch Übergangsflächen zwischen verschiedenen Systemhöhen und beziehen sich immer auf die Soll-Gleislage. Die tatsächliche Bahnsteighöhe kann auf Grund der Ist-Gleislage und Instandhaltungstoleranzen von der Systemhöhe abweichen.", identity],
    "ALG_REISENDENAUFKOMM": ["Reisendenaufkommen", "Angabe des Reisendenaufkommens.", identity],
    "ALG_SERVICEEINRICHT": ["Serviceeinrichtung", "Übersicht der Serviceeinrichtungen in der Betriebsstelle.", identity],
    "ALG_SOEZ_DETAILS_STELLE": ["Streckenöffnungszeiten", "Link zur Anzeige der täglichen Streckenöffnungszeiten für das laufende Fahrplanjahr", identity],
    "ALG_STAAT": ["Staat", "Mitgliedstaat der europäischen Union (+UK +NO +CH).", identity],
    "ALG_STRECKE_DET": [],
    "ALG_TUNNEL_DET": [],
    "ALG_ZZS_BST": [],
    "BST_HS_DS_KZ": [],
    "BST_RL100": ["RL100", "Angabe der Abkürzung für Betriebsstellen nach Richtlinie 100.", identity],
    "BST_STELLENART": ["Betriebsstellenart", "Art der Einrichtung hinsichtlich der vorherrschenden betrieblichen Funktionen.", identity],
    "BST_STELLE_NAME": ["Betriebsstellen Name", "In der Regel auf die betreffende Ansiedlung (Stadt/Dorf) oder auf verkehrsbetriebliche Zwecke bezogene Bezeichnung.", identity],
    "BST_TAF_TAP_PC": ["TAF/TAP primary Code", "Für TAF/TAP entwickelter Primärcode (siehe auch https://crd.tsi-cc.eu/CRD/Login.action).", identity],
    "DET_GLEISNUTZLAENGEN": [],
    "DET_GLEISZUORDNUNG": [],
    "HAS_BSZ": [],
    "HAUPTLAGE": ["Hauptlage", "Teil der Hauptlage.", identity],
    "ID": ["Identifikatinsnummer", "Identifkitationsnummer im Infrastrukturregister.", identity],
    "JAHR": ["Jahr", "Angabe des Jahres.", identity],
    "LADE_ID": ["Lade ID", "Lade Identifkationsnummer im Infrastrukturregister.", identity],
    "LAGE_KM_I": ["Streckenlage (DB codiert)", "Lage der Betriebsstelle auf der zugeordneten Strecke als Kilometerierung.", str_to_int],
    "LAGE_KM_V": ["Streckenlage [km]", "Lage der Betriebsstelle auf der zugeordneten Strecke als Kilometerierung.", db_km_to_km],
    "STEL_ID": ["Stellen ID", "Stellen Identifkationsnummer im Infrastrukturregister.", identity],
    "STRNR": ["Strecken-Nr.", "Bezeichnung und Identifizierung für Eisenbahnstrecken in Deutschland.", identity],
}

""" Übersicht: Betriebsstellenarten
    Abzw
        Abzweigstelle
    Anst
        Anschlussstelle
    Awanst
        Ausweichanschlussstelle
    Bf
        Bahnhof
    Bft
        Bahnhofsteil
    Bk
        Blockstelle
    Bush
        Bushaltestelle
    Dkst
        Deckungsstelle
    Est
        Einsatzstelle für Zugpersonal
    Fwst
        Fernwirkstelle
    Gp
        Grenzpunkt
    Hp
        Haltepunkt
    LGr
        Landesgrenze
    LW
        Laufweg
    Museum
        Museumsbahnhof
    PDGr
        Produktionsdurchführungsgrenze
    RBGr
        Regionalbereichsgrenze
    Sbk
        Selbsttätige Blockstelle
    Schstr
        Schutzstrecke
    Slst
        Schiffslandestelle
    Sp
        Schaltposten
    Strw
        Streckenwechsel
    Stub
        Streckenübergang
    Tp
        Tarifpunkt
    Urw
        Umrichterwerk
    Uw
        Unterwerk
    Üst
        Überleitstelle
    Zes
        Zentralschaltstelle
    NE-*
        Nichtbundeseigene Eisenbahn
    vp-*
        verpachtet
"""


### STRECKENÜBERGÄNGE

ISR_PROPERTIES_JUNCTION = {
    "ALG_DBNETZ_STRECKE": [],
    "BST_STELLENART": ["Betriebsstellenart", "Art der Einrichtung hinsichtlich der vorherrschenden betrieblichen Funktionen.", identity],
    "BST_STELLE_NAME": ["Betriebsstellen Name", "In der Regel auf die betreffende Ansiedlung (Stadt/Dorf) oder auf verkehrsbetriebliche Zwecke bezogene Bezeichnung.", identity],
    "BST_TAF_TAP_PC": ["TAF/TAP primary Code", "Für TAF/TAP entwickelter Primärcode (siehe auch https://crd.tsi-cc.eu/CRD/Login.action).", identity],
    "ID": ["Identifikatinsnummer", "Identifkitationsnummer im Infrastrukturregister.", identity],
    "JAHR": ["Jahr", "Angabe des Jahres.", identity],
    "LADE_ID": ["Lade ID", "Lade Identifkationsnummer im Infrastrukturregister.", identity],
    "REF_RL100": ["Referenzierte RL100", "Angabe der Abkürzung der dem Streckenübergang zugeordneten Betriebsstelle nach Richtlinie 100.", identity],
    "STEL_ID": ["Stellen ID", "Stellen Identifkationsnummer im Infrastrukturregister.", identity],
    "STRECKE1": ["Strecken-Nr. 1", "Die erste Streckennummer, die dem Streckenübergang zugeordnet ist.", identity],
    "STRECKE2": ["Strecken-Nr. 2", "Die zweite Streckennummer, die dem Streckenübergang zugeordnet ist.", identity],
}


### TUNNEL

ISR_PROPERTIES_TUNNEL = {
    "ALG_BESCHR_DOC": ["Tunneldetails", "Dokument mit Details zum Tunnel.", identity],
    "ALG_DBNETZ_STRECKE": [],
    "ALG_DIESEL_ERL": ["Diesel- oder andere Verbrennungsantriebe zulässig", "Angabe, ob der Einsatz von Diesel- oder anderen Verbrennungsantrieben im Tunnel erlaubt ist.", identity],
    "ALG_EG_PRUEF_TUNN": ["EG-Prüferklärung für Tunnel (SRT)", "Eindeutige Nummer der EG-Erklärung gemäß den Formatvorgaben in ?Document about practical arrangements for transmitting interoperability documents'.", identity],
    "ALG_ERF_BRANDKAT_FZG": ["Erforderliche Brandkategorie von Fahrzeugen", "Kategorisierung, wie ein Reisezug bei einem Brand im Zug für einen definierten Zeitraum weiter betrieben werden kann.", identity],
    "ALG_ERF_BRANDKAT_NAT": ["Erforderliche nationale Brandkategorie", "Kategorisierung, wie ein Reisezug bei einem Brand im Zug für einen definierten Zeitraum gemäß etwaigen nationalen Vorschriften weiter betrieben werden kann.", identity],
    "ALG_IM_CODE": ["Infrastrukturbetreiber Code", "Numerische Codierung des Infrastrukturbetreibers gemäß UIC-Liste (verfügbar unter http://www.uic.org -> Company Codes) z. B. 0080 für die DB Netz AG.", identity],
    "ALG_LICHT_QUERSCHN": ["Lichter Querschnitt [m^2]", "Kleinster tatsächlicher Querschnitt des Tunnels in Quadratmeter.", str_to_float],
    "ALG_NOTFALLPLAN_VORH": ["Notfallplan vorhanden", "Angabe, ob eine Notfallplan vorhanden ist.", identity],
    "ALG_TSI_KONF": ["Konformität des Tunnels mit der TSI INF", "Konformität des Tunnels mit der TSI INF bei zulässiger Höchstgeschwindigkeit.", identity],
    "ALG_TUNNELANFANG": ["Tunnelanfang", "Geografische Koordinaten in Dezimalgrad am Beginn eines Tunnels.", str_to_lat_lon],
    "ALG_TUNNELART": ["Tunnelart", "Angabe, ob der Tunnel ein- oder mehrgleisig ist.", identity],
    "ALG_TUNNELENDE": ["Tunnelende", "Geografische Koordinaten in Dezimalgrad am Ende eines Tunnels.", str_to_lat_lon],
    "ALG_TUNNELLAENGE": ["Tunnellänge [m]", "Länge des Tunnels in Metern von der Tunneleinfahrt bis zur Tunnelausfahrt, (siehe auch Anlage 3 Längenabhängige Darstellung der Tunnel unter http://fahrweg.dbnetze.com/fahrweg-de/nutzungsbedingungen/infrastrukturregister/grundsaetze.html).", str_to_float],
    "ALG_TUNNELNAME": ["Name des Tunnels", "Angabe des Tunnelnamens.", identity],
    "ALG_TUNN_DETAILS": ["Tunneldetails", "Dokument mit Details zum Tunnel.", identity],
    "ALG_TUNN_NOTAUS_KZ": ["Kennzeichnung von Notausgängen im Tunnel", "Dokument zur Beschreibung der Notausgänge in Tunneln.", identity],
    "BAUWERKSID": ["Bauwerksidentifikationsnummer", "Identifikationsnummer des Bauwerks.", identity],
    "DET_STR_NR": [],
    "ID": ["Identifikatinsnummer", "Identifkitationsnummer im Infrastrukturregister.", identity],
    "JAHR": ["Jahr", "Angabe des Jahres.", identity],
    "KMBIS_I": ["Kilometer (Tunnelende) - DB Codierung", "Angaben der Kilometrierung (metergenau) des Tunnelendes (DB Codierung).", db_km_alt_to_km],
    "KMVON_I": ["Kilometer (Tunnelanfang) - DB Codierung", "Angaben der Kilometrierung (metergenau) des Tunnelanfangs (DB Codierung).", db_km_alt_to_km],
    "LADE_ID": ["Lade ID", "Lade Identifkationsnummer im Infrastrukturregister.", identity],
    "STEL_ID": ["Stellen ID", "Stellen Identifkationsnummer im Infrastrukturregister.", identity],
    "STRNR": ["Strecken-Nr.", "Bezeichnung und Identifizierung für die Strecke, die dem Tunnel zugeordnet ist.", identity],
    "TUNN_ID": ["Tunnelidentifikationsnummer", "Identifikationsnummer des Tunnels.", identity],
}


##### OSM - OpenStreetMap


# fmt: on
